-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with MAST.Systems,MAST.Transactions,Mast.Processing_Resources,
  Mast.Processing_Resources.FP,Mast.Timers,Mast.Drivers,
  Ada.Text_IO,Mast.Linear_Translation, Mast.Max_Numbers;
use Ada.Text_IO;
use type Mast.Timers.System_Timer_Ref;

package body Mast.Linear_Analysis_Tools is

   type Processor_ID is new Natural;
   type Transaction_ID is new Natural;
   type Task_ID is new Natural;
   type Long_Int is range -(2**31-1)..2**31-2;
   -- This is because the compiler seems to fail with a Long_Integer

   ------------
   -- Modulus --
   ------------

   function Modulus(A,B:Time) return Time is
   begin
      return A-Time'Floor(A/B)*B;
   end Modulus;

   ------------
   -- Ceil0 --
   ------------

   function Ceil0(X:Time) return Time is
   begin
      if X<0.0 then return 0.0;
      else return Time'Ceiling(X);
      end if;
   end Ceil0;

   -----------------------
   -- Holistic_Analysis --
   -----------------------

   procedure Holistic_Analysis
     (The_System : in out MAST.Systems.System;
      Verbose : in Boolean:=True)
   is
      Max_Processors:constant Processor_ID:=Processor_ID
        (Processing_Resources.Lists.Size
         (The_System.Processing_Resources));
      Max_Transactions:constant Transaction_ID:=
        Transaction_Id((Mast.Max_Numbers.Calculate_Max_Transactions
                        (The_System)));
      Max_Tasks_Per_Transaction:constant Task_ID:=Task_ID
        (Mast.Max_Numbers.Calculate_Max_Tasks_Per_Transaction(The_System));

      subtype Processor_ID_Type is Processor_ID
        range 0..Max_Processors;
      subtype Transaction_ID_Type is Transaction_ID
        range 0..Max_Transactions;
      subtype Task_ID_Type is Task_ID
        range 0..Max_Tasks_Per_Transaction;

      package Translation is new Linear_Translation
        (Processor_ID_Type, Transaction_ID_Type, Task_ID_Type,
         Max_Processors, Max_Transactions, Max_Tasks_Per_Transaction);

      use Translation;

      --------------
      -- WTindell --
      --------------

      function WTindell
        (Ta:Transaction_ID_Type;
         K:Task_ID_Type;
         Q:Long_Int;
         Transaction : in Linear_Transaction_System)
        return Time
      is
         Wc,Wcant,Jitter:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
         Done:Boolean;
      begin
         Proc:=Transaction(Ta).The_Task(K).Procij;
         Prio:=Transaction(Ta).The_Task(K).Prioij;
         Wcant:=Time(Q)*Transaction(Ta).The_Task(K).Cijown
           +Transaction(Ta).The_Task(K).Bij;
         loop
            Wc:=Time(Q)*Transaction(Ta).The_Task(K).Cijown
              +Transaction(Ta).The_Task(K).Bij;

            for I in Transaction_ID_Type range 1..Max_Transactions
            loop
               for J in 1..Transaction(I).Ni
               loop
                  if (Transaction(I).The_Task(J).Procij=Proc)
                    and (Transaction(I).The_Task(J).Prioij>=Prio)
                    and not((I=Ta)and(J=K))
                  then
                     if Transaction(I).The_Task(J).Model=Unbounded_Effects then
                        return Large_Time;
                     elsif
                       Transaction(I).The_Task(J).Jitter_Avoidance then
                        Jitter:=0.0;
                     else
                        Jitter:=Transaction(I).The_Task(J).Jij;
                     end if;
                     Wc:=Wc+Time'Ceiling((Wcant+Jitter)/
                                         Transaction(I).The_Task(J).Tij)*
                       Transaction(I).The_Task(J).Cij;
                  end if;
               end loop;
            end loop;
            Done:=Wc=Wcant;
            Wcant:=Wc;
            exit when Done;
         end loop;
         return Wc;
      end WTindell;

      Q:Long_Int;
      R_Ij,W_Ij,Rmax:Time;
      Acum,Acumb:Time;
      Done:Boolean;
      Transaction : Linear_Transaction_System;

   begin
      if Verbose then
         New_Line;Put_Line("Holistic Analysis");
      end if;
      Translate_Linear_System(The_System,Transaction,Verbose);
      --show_linear_translation(Transaction);
      for I in 1..Max_Transactions
      loop
         Acum:=0.0;
         Acumb:=0.0;
         for L in 1..Transaction(I).Ni loop
            Acum:=Acum+Transaction(I).The_Task(L).Jinit;
            Transaction(I).The_Task(L).Jij:=Acum-Acumb;
            Acum:=Acum+Transaction(I).The_Task(L).Cijown;
            Transaction(I).The_Task(L).Rij:=Acum;
            Acumb:=Acumb+Transaction(I).The_Task(L).Cbijown;
            Transaction(I).The_Task(L).Rbij:=Acumb;
         end loop;
      end loop;

      loop
         Done:=True;
         for I in 1..Max_Transactions
         loop
            for J in 1..Transaction(I).Ni
            loop
               if Transaction(I).The_Task(J).Model /= Unbounded_Response
                 and Transaction(I).The_Task(J).Model /= Unbounded_Effects then
                  Q:=0;Rmax:=0.0;
                  loop
                     Q:=Q+1;
                     W_Ij:=WTindell(I,J,Q,Transaction);
                     R_Ij:=W_Ij+Transaction(I).The_Task(J).Jij-Time(Q-1)*
                       Transaction(I).The_Task(J).Tijown;
                     if J>1 then
                        R_Ij:=R_Ij+Transaction(I).The_Task(J-1).Rbij;
                     end if;
                     if R_Ij>Rmax then Rmax:=R_Ij; end if;
                     if Rmax>Analysis_Bound then
                        Done:=False;
                        for K in J..Transaction(I).Ni loop
                           Transaction(I).The_Task(K).Model:=Unbounded_Effects;
                           Transaction(I).The_Task(K).Rij:=Large_Time;
                        end loop;
                     end if;
                     exit when
                       (W_Ij<=Time(Q)*Transaction(I).The_Task(J).Tijown)
                       or Transaction(I).The_Task(J).Model=Unbounded_Effects;
                  end loop;
                  if Rmax>Transaction(I).The_Task(J).Rij
                  then
                     Done:=False;
                     Transaction(I).The_Task(J).Rij:=Rmax;
                     if J<Transaction(I).Ni then
                        Transaction(I).The_Task(J+1).Jij:=
                          Rmax-Transaction(I).The_Task(J).Rbij;
                     end if;
                  end if;
               end if;
            end loop;
         end loop;
         exit when Done;
      end loop;

      Translate_Linear_Analysis_Results(Transaction,The_System);
   end Holistic_Analysis;


   procedure Offset_Based_Unoptimized_Analysis
     (The_System : in out MAST.Systems.System;
      Verbose : in Boolean:=True)
   is

      Max_Processors:constant Processor_ID:=Processor_ID
        (Processing_Resources.Lists.Size
         (The_System.Processing_Resources));
      Max_Transactions:constant Transaction_ID:=
        Transaction_Id((Mast.Max_Numbers.Calculate_Max_Transactions
                        (The_System)));
      Max_Tasks_Per_Transaction:constant Task_ID:=Task_ID
        (Mast.Max_Numbers.Calculate_Max_Tasks_Per_Transaction(The_System));

      subtype Processor_ID_Type is Processor_ID
        range 0..Max_Processors;
      subtype Transaction_ID_Type is Transaction_ID
        range 0..Max_Transactions;
      subtype Task_ID_Type is Task_ID
        range 0..Max_Tasks_Per_Transaction;

      package Translation is new Linear_Translation
        (Processor_ID_Type, Transaction_ID_Type, Task_ID_Type,
         Max_Processors, Max_Transactions,Max_Tasks_Per_Transaction);

      use Translation;

      ----------
      -- Fijk --
      ----------

      function Fijk
        (I:Transaction_ID_Type;
         J,K:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time
      is
      begin
         return Transaction(I).The_Task(J).Tij-
           Modulus(Transaction(I).The_Task(K).Oij+
                   Transaction(I).The_Task(K).Jij-
                   Transaction(I).The_Task(J).Oij,
                   Transaction(I).The_Task(J).Tij);
      end Fijk;


      ----------
      -- Wik --
      ----------

      function Wik
        (I:Transaction_ID_Type;
         K:Task_ID_Type;
         T:Time;
         A:Transaction_ID_Type;
         B:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time
      is
         Acum:Time:=0.0;
         Jitter:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
      begin
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         for J in 1..Transaction(I).Ni
         loop
            if (Transaction(I).The_Task(J).Procij=Proc)
              and (Transaction(I).The_Task(J).Prioij>=Prio)
              and not((I=A)and(J=B))
            then
               if Transaction(I).The_Task(J).Model=Unbounded_Effects then
                  return Large_Time;
               elsif Transaction(I).The_Task(J).Independent then
                  if Transaction(I).The_Task(J).Jitter_Avoidance then
                     Jitter:=0.0;
                  else
                     Jitter:=Transaction(I).The_Task(J).Jij;
                  end if;
                  Acum:=Acum+Time'Ceiling((T+Jitter)/
                                          Transaction(I).The_Task(J).Tij)*
                    Transaction(I).The_Task(J).Cij;

               elsif not Transaction(I).The_Task(K).Independent then
                  Acum:=Acum+(Time'Floor
                              ((Transaction(I).The_Task(J).Jij+
                                Fijk(I,J,K,Transaction))/
                               Transaction(I).The_Task(J).Tij)+
                              Time'Ceiling((T-Fijk(I,J,K,Transaction))/
                                           Transaction(I).The_Task(J).Tij))*
                    Transaction(I).The_Task(J).Cij;
               end if;
            end if;
         end loop;
         return Acum;
      end Wik;

      ----------
      -- W_DO --
      ----------

      function W_DO
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         Q:Long_Int;
         Transaction : in Linear_Transaction_System) return Time
      is
         Wc,Wcant,Wc_Ik,MaxWik:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
         Done:Boolean;
      begin
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         Wcant:=Time(Q)*Transaction(A).The_Task(B).Cijown
           +Transaction(A).The_Task(B).Bij;
         loop
            Wc:=Time(Q)*Transaction(A).The_Task(B).Cijown
              +Transaction(A).The_Task(B).Bij
              +Wik(A,C,Wcant,A,B,Transaction);
            for I in 1..Max_Transactions
            loop
               if I/=A then
                  MaxWik:=0.0;
                  for K in 1..Transaction(I).Ni
                  loop
                     Wc_IK:=0.0;
                     if (Transaction(I).The_Task(K).Procij=Proc)
                       and(Transaction(I).The_Task(K).Prioij>=Prio)
                       and not((I=A)and(K=B))
                     then
                        if Transaction(I).The_Task(K).Model=Unbounded_Effects
                        then
                           return Large_Time;
                        else
                           Wc_Ik:=Wc_IK+Wik(I,K,Wcant,A,B,Transaction);
                        end if;

                        if Wc_Ik>MaxWik then MaxWik:=Wc_Ik;end if;
                     end if;
                  end loop;
                  Wc:=Wc+MaxWik;
               end if;
            end loop;
            Done:=(Wc=Wcant);
            Wcant:=Wc;
            exit when Done;
         end loop;
         return Wc;
      end W_DO;

      P,P0:Long_Int;
      R_Abc,W_Abc,Rmax:Time;
      Acum,Acumb:Time;
      Done:Boolean;

      Transaction : Linear_Transaction_System;

   begin
      if Verbose then
         New_Line;Put_Line("Offset Based Unoptimized Analysis");
      end if;
      Translate_Linear_System(The_System,Transaction,Verbose);
      --show_linear_translation(Transaction);
      for A in 1..Max_Transactions
      loop
         Acum:=0.0;
         Acumb:=0.0;
         for B in 1..Transaction(A).Ni
         loop
            Acum:=Acum+Transaction(A).The_Task(B).Jinit;
            Transaction(A).The_Task(B).Jij:=Acum-Acumb;
            Acum:=Acum+Transaction(A).The_Task(B).Cijown;
            Transaction(A).The_Task(B).Rij:=Acum;
            Transaction(A).The_Task(B).Oij:=Acumb;
            if Transaction(A).The_Task(B).Oij >Transaction(A).The_Task(B).Tij
              and (Transaction(A).Kind_Of_Event=Sporadic or
               Transaction(A).Kind_Of_Event=Bursty) then
               for K in B..Transaction(A).Ni loop
                  Transaction(A).The_Task(B).Independent:=True;
               end loop;
            end if;
            Acumb:=Acumb+Transaction(A).The_Task(B).Cbijown;
            Transaction(A).The_Task(B).Rbij:=Acumb;
         end loop;
      end loop;

      loop
         Done:=True;
         for A in 1..Max_Transactions
         loop
            for B in 1..Transaction(A).Ni
            loop
               if Transaction(A).The_Task(B).Model /= Unbounded_Response
                 and Transaction(A).The_Task(B).Model /= Unbounded_Effects then
                  Rmax:=0.0;
                  for C in 1..Transaction(A).Ni
                  loop
                     if (Transaction(A).The_Task(C).Prioij>=
                         Transaction(A).The_Task(B).Prioij)
                       and (Transaction(A).The_Task(C).Procij=
                            Transaction(A).The_Task(B).Procij)
                       and  Transaction(A).The_Task(B).Model /=
                       Unbounded_Effects
                     then
                        P0:=-Long_Int
                          (Time'Floor((Transaction(A).The_Task(B).Jij+
                                       Fijk(A,B,C,Transaction))/
                                      Transaction(A).The_Task(B).Tijown))+1;
                        P:=P0-1;
                        loop
                           P:=P+1;
                           W_Abc:=W_DO(A,B,C,P-P0+1,Transaction);
                           if Transaction(A).The_Task(B).Independent then
                               R_Abc:=Time(P-P0)*Transaction(A).The_Task(B).Tijown+W_Abc+
                                 Transaction(A).The_Task(B).Oij+
                                 Transaction(A).The_Task(B).Jij;
                           else
                             R_Abc:=-Fijk(A,B,C,Transaction)-Time(P-1)*
                               Transaction(A).The_Task(B).Tijown+W_Abc+
                               Transaction(A).The_Task(B).Oij;
                           end if;
                           if R_Abc>Rmax then Rmax:=R_Abc;end if;
                           if Rmax>Analysis_Bound then
                              Done:=False;
                              for K in B..Transaction(A).Ni loop
                                 Transaction(A).The_Task(K).Model:=
                                   Unbounded_Effects;
                                 Transaction(A).The_Task(K).Rij:=Large_Time;
                              end loop;
                           end if;
                           exit when R_Abc<=Transaction(A).The_Task(B).Tijown+
                             Transaction(A).The_Task(B).Oij
                             or Transaction(A).The_Task(B).Model=
                             Unbounded_Effects;
                        end loop;
                     end if;
                  end loop;
                  if Rmax>Transaction(A).The_Task(B).Rij then
                     Done:=False;Transaction(A).The_Task(B).Rij:=Rmax;
                     if B<Transaction(A).Ni then
                        Transaction(A).The_Task(B+1).Jij:=
                          Transaction(A).The_Task(B).Rij-
                          Transaction(A).The_Task(B).Rbij;
                     end if;
                  end if;
               end if;
            end loop;
         end loop;
         exit when Done;
      end loop;

      Translate_Linear_Analysis_Results(Transaction,The_System);
   end Offset_Based_Unoptimized_Analysis;



   procedure Offset_Based_Analysis
     (The_System : in out MAST.Systems.System;
      Verbose : in Boolean:=True)
   is

      Max_Processors:constant Processor_ID:=Processor_ID
        (Processing_Resources.Lists.Size
         (The_System.Processing_Resources));
      Max_Transactions:constant Transaction_ID:=
        Transaction_Id((Mast.Max_Numbers.Calculate_Max_Transactions
                        (The_System)));
      Max_Tasks_Per_Transaction:constant Task_ID:=Task_ID
        (Mast.Max_Numbers.Calculate_Max_Tasks_Per_Transaction(The_System));

      subtype Processor_ID_Type is Processor_ID
        range 0..Max_Processors;
      subtype Transaction_ID_Type is Transaction_ID
        range 0..Max_Transactions;
      subtype Task_ID_Type is Task_ID
        range 0..Max_Tasks_Per_Transaction;

      package Translation is new Linear_Translation
        (Processor_ID_Type, Transaction_ID_Type, Task_ID_Type,
         Max_Processors, Max_Transactions,Max_Tasks_Per_Transaction);

      use Translation;

      function FTijk
        (I:Transaction_ID_Type;
         J,K:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time is
      begin
         return Transaction(I).The_Task(K).Tij-
           Modulus(Transaction(I).The_Task(K).Oij+
                   Transaction(I).The_Task(K).Jij,Transaction(I).The_Task(K).Tij)+
           Transaction(I).The_Task(J).Oij;
      end FTijk;

      ------------
      -- Same_H --
      ------------

      function Same_H
        (I:Transaction_ID_Type;
         J,K:Task_ID_Type;
         Prio:Priority;
         Transaction : in Linear_Transaction_System) return Boolean
      is
         Jj,Kk:Task_ID_Type;
      begin
         if Transaction(I).The_Task(J).Independent or
           Transaction(I).The_Task(K).Independent then return False;
         end if;
         if J>K then Jj:=K;Kk:=J; else Jj:=J;Kk:=K;end if;
         for L in Jj..Kk
         loop
            if (Transaction(I).The_Task(L).Procij=
                Transaction(I).The_Task(Jj).Procij)
              and(Transaction(I).The_Task(L).Prioij<Prio)
            then
               return False;
            end if;
         end loop;
         return True;
      end Same_H;

      -----------
      -- In_MP --
      -----------

      function In_MP
        (I:Transaction_ID_Type;
         J:Task_ID_Type;
         Prio:Priority;
         Proc:Processor_ID_Type;
         Transaction : in Linear_Transaction_System) return Boolean
      is
      begin
         if Transaction(I).The_Task(J).Independent
          and(Transaction(I).The_Task(J).Prioij>=Prio) then
            return True;
         end if;
         for L in 1..J loop
            if (Transaction(I).The_Task(L).Procij=Proc)
              and(Transaction(I).The_Task(L).Prioij<Prio)
            then
               return False;
            end if;
         end loop;
         return True;
      end In_MP;

      -----------------------
      -- Resolve_Conflicts --
      -----------------------

      function Resolve_Conflicts
        (I:Transaction_ID_Type;
         K:Task_ID_Type;
         T:Time;
         A:Transaction_ID_Type;
         B:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time
      is
         The_First_P,P0,P0kk:Long_Int;
         Proc:Processor_ID_Type;
         Prio:Priority;
         J:Task_ID_Type;
         Total,PLinkial,Sum,Cell:Time;
      begin
         Total:=0.0;
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         J:=Transaction(I).Ni;
         loop
            exit when (Transaction(I).The_Task(J).Procij=Proc)
              and(Transaction(I).The_Task(J).Prioij>=Prio);
            J:=J-1;
         end loop;

         The_First_P:=Long_Int
           (-Time'Floor
            ((Transaction(I).The_Task(J).Jij+FTijk(I,J,K,Transaction))
             /Transaction(I).The_Task(K).Tij)+Time'(1.0));
         P0kk:=Long_Int(-Time'Floor((Transaction(I).The_Task(K).Jij+
                                     FTijk(I,K,K,Transaction))
                                    /Transaction(I).The_Task(K).Tij)+Time'(1.0));
         for P in The_First_P..0
         loop
            PLinkial:=0.0;
            Sum:=0.0;
            for J in 1..Transaction(I).Ni
            loop
               if Transaction(I).The_Task(J).Procij=Proc then
                  if Transaction(I).The_Task(J).Prioij>=Prio then
                     Cell:=0.0;
                     P0:=Long_Int
                       (-Time'Floor
                        ((Transaction(I).The_Task(J).Jij+
                          FTijk(I,J,K,Transaction))
                         /Transaction(I).The_Task(K).Tij)+Time'(1.0));
                     if (P>=P0)and
                       (T>FTijk(I,J,K,Transaction)+
                        Time(P-1)*Transaction(I).The_Task(K).Tij)
                     then
                        Cell:=Transaction(I).The_Task(J).Cij;
                     end if;
                     if (J>K and P>=P0kk and not
                         Same_H(I,J,K,Prio,Transaction))
                       or Transaction(I).The_Task(J).Independent
                     then
                        Cell:=0.0;
                     end if;
                     Sum:=Sum+Cell;
                  else
                     Sum:=0.0;
                  end if;
                  if Sum>PLinkial then PLinkial:=Sum;end if;
               end if;
            end loop;
            Total:=Total+PLinkial;
         end loop;
         return Total;
      end;

      ----------------------------
      -- Resolve_Self_Conflicts --
      ----------------------------

      function Resolve_Self_Conflicts
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         T:Time;
         Pa:Long_Int;
         Transaction : in Linear_Transaction_System) return Time
      is
         The_First_P,P0,P0cc:Long_Int;
         Proc:Processor_ID_Type;
         Prio:Priority;
         J:Task_ID_Type;
         Total,PLinkial,Sum,Cell:Time;
      begin
         Total:=0.0;
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         J:=Transaction(A).Ni;
         loop
            exit when (Transaction(A).The_Task(J).Procij=Proc) and
              (Transaction(A).The_Task(J).Prioij>=Prio);
            J:=J-1;
         end loop;
         The_First_P:=Long_Int
           (-Time'Floor((Transaction(A).The_Task(J).Jij+
                         FTijk(A,J,C,Transaction))
                        /Transaction(A).The_Task(C).Tij)+Time'(1.0));
         P0cc:=Long_Int
           (-Time'Floor((Transaction(A).The_Task(C).Jij+
                         FTijk(A,C,C,Transaction))
                        /Transaction(A).The_Task(C).Tij)+Time'(1.0));
         for P in The_First_P..0
         loop
            PLinkial:=0.0;
            Sum:=0.0;
            for J in 1..Transaction(A).Ni
            loop
               if Transaction(A).The_Task(J).Procij=Proc then
                  if Transaction(A).The_Task(J).Prioij>=Prio then
                     Cell:=0.0;
                     P0:=Long_Int
                       (-Time'Floor((Transaction(A).The_Task(J).Jij+
                                     FTijk(A,J,C,Transaction))
                                    /Transaction(A).The_Task(C).Tij)+Time'(1.0));
                     if (P>=P0)and
                       (T>FTijk(A,J,C,Transaction)+
                        Time(P-1)*Transaction(A).The_Task(C).Tij)
                     then
                        Cell:=Transaction(A).The_Task(J).Cij;
                     end if;
                     if (J>C) and (P>=P0cc) and not
                       Same_H(A,J,C,Prio,Transaction)
                     then
                        Cell:=0.0;
                     end if;
                     if (J<C) and (P<=Pa) and not
                       Same_H(A,J,C,Prio,Transaction)
                     then
                        Cell:=0.0;
                     end if;
                     if ((J>=B)and(P>Pa)) or ((J>B)and(P=Pa)) then
                        Cell:=0.0;
                     end if;
                     if Transaction(A).The_Task(J).Independent then
                        Cell:=0.0;
                     end if;
                     Sum:=Sum+Cell;
                  else
                     Sum:=0.0;
                  end if;
                  if Sum>PLinkial then
                     PLinkial:=Sum;
                  end if;
               end if;
            end loop;
            Total:=Total+PLinkial;
         end loop;
         return Total;
      end;

      ------------
      -- Wik_LH --
      ------------

      function Wik_LH
        (I:Transaction_ID_Type;
         K:Task_ID_Type;
         T:Time;
         A:Transaction_ID_Type;
         B:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time
      is
         Acum,Jitter:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
      begin
         if Transaction(I).The_Task(K).Independent then
            Acum:=0.0;
         else
            Acum:=Resolve_Conflicts(I,K,T,A,B,Transaction);
         end if;
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         for J in 1..Transaction(I).Ni
         loop
            if (Transaction(I).The_Task(J).Procij=Proc) and
              In_MP(I,J,Prio,Proc,Transaction)
            then
               if Transaction(I).The_Task(J).Model=Unbounded_Effects then
                  return Large_Time;
               elsif Transaction(I).The_Task(J).Independent then
                  if Transaction(I).The_Task(J).Jitter_Avoidance then
                     Jitter:=0.0;
                  else
                     Jitter:=Transaction(I).The_Task(J).Jij;
                  end if;

                  if I=A and K=B then
                     Acum:=Acum+Time'Ceiling((T+Jitter)/
                                             Transaction(I).The_Task(J).Tijown)*
                       Transaction(I).The_Task(J).Cijown;
                  else
                     Acum:=Acum+Time'Ceiling((T+Jitter)/
                                             Transaction(I).The_Task(J).Tij)*
                       Transaction(I).The_Task(J).Cij;
                  end if;

               elsif not Transaction(I).The_Task(K).Independent then
                  Acum:=Acum+Ceil0((T-FTijk(I,J,K,Transaction))/
                                   Transaction(I).The_Task(J).Tij)*
                    Transaction(I).The_Task(J).Cij;
               end if;
            end if;

         end loop;
         return Acum;
      end Wik_LH;

      ------------
      -- Wac_LH --
      ------------

      function Wac_LH
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         T:Time;
         P:Long_Int;
         Transaction : in Linear_Transaction_System) return Time
      is
         Acum,Post,Jitter:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
      begin
         if Transaction(A).The_Task(C).Independent then
            Acum:=0.0;
         else
            Acum:=Resolve_Self_Conflicts(A,B,C,T,P,Transaction);
         end if;

         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         Post:=Time(P)*Transaction(A).The_Task(B).Cijown;
         for J in 1..Transaction(A).Ni
         loop
            if (Transaction(A).The_Task(J).Procij=Proc) and
              In_MP(A,J,Prio,Proc,Transaction)
            then
               if Transaction(A).The_Task(J).Model=Unbounded_Effects then
                  return Large_Time;
               elsif Transaction(A).The_Task(J).Independent then
                  if Transaction(A).The_Task(J).Jitter_Avoidance then
                     Jitter:=0.0;
                  else
                     Jitter:=Transaction(A).The_Task(J).Jij;
                  end if;
                  if J/=B then
                   Acum:=Acum+Time'Ceiling((T+Jitter)/
                                          Transaction(A).The_Task(J).Tij)*
                    Transaction(A).The_Task(J).Cij;
                  end if  ;
               else
                  if J<B then
                     Acum:=Acum+Ceil0((T-FTijk(A,J,C,Transaction))/
                                      Transaction(A).The_Task(J).Tij)*
                       Transaction(A).The_Task(J).Cij;
                  end if;
                  if J>B then
                     Post:=Post+Time'Min
                       (Time(P-1),Ceil0((T-FTijk(A,J,C,Transaction))/
                                        Transaction(A).The_Task(J).Tij))*
                       Transaction(A).The_Task(J).Cij;
                  end if;
               end if;
            end if;
         end loop;
         return Acum+Time'Max(0.0,Post);
      end Wac_LH;

      -------------
      -- W_DO_LH --
      -------------

      function W_DO_LH
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         P:Long_Int;
         Transaction : in Linear_Transaction_System) return Time
      is
         Wc,Wcant,Wc_Ik,MaxWik:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
         Done:Boolean;
         P0:Long_Int;
      begin
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         if Transaction(A).The_Task(B).Independent then
            P0:=1;
         else
             P0:=-Long_Int(Time'Floor
                       (Transaction(A).The_Task(B).Jij+
                        FTijk(A,B,C,Transaction))/
                       Transaction(A).The_Task(B).Tij)+1;
         end if;
         Wcant:=Time(P-P0+1)*Transaction(A).The_Task(B).Cijown
           +Transaction(A).The_Task(B).Bij;
         loop
            Wc:=Wac_LH(A,B,C,Wcant,P,Transaction)
              +Transaction(A).The_Task(B).Bij;
            for I in 1..Max_Transactions
            loop
               if I/=A then
                  MaxWik:=0.0;
                  for K in 1..Transaction(I).Ni
                  loop
                     if (Transaction(I).The_Task(K).Procij=Proc)
                       and(Transaction(I).The_Task(K).Prioij>=Prio)
                       and not((I=A)and(K=B))
                       and
                       ((K=1)or else
                        (Transaction(I).The_Task(K-1).Procij/=Proc)
                        or else(Transaction(I).The_Task(K-1).Prioij<Prio))
                     then
                        Wc_Ik:=Wik_LH(I,K,Wcant,A,B,Transaction);
                        if Wc_Ik>MaxWik then
                           MaxWik:=Wc_Ik;
                        end if;
                     end if;
                  end loop;
                  Wc:=Wc+MaxWik;
               end if;
            end loop;
            Done:=(Wc=Wcant);
            Wcant:=Wc;
            exit when Done;
         end loop;
         return Wc;
      end W_DO_LH;

      -----------------
      -- Busy_Period --
      -----------------

      function Busy_Period
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Time
      is
         Wc,Wcant,Wc_Ik,MaxWik:Time;
         Prio:Priority;
         Proc:Processor_ID_Type;
         Done:Boolean;
      begin
         Proc:=Transaction(A).The_Task(B).Procij;
         Prio:=Transaction(A).The_Task(B).Prioij;
         Wcant:=Transaction(A).The_Task(B).Cijown
           +Transaction(A).The_Task(B).Bij;
         loop
            Wc:=Wik_LH(A,C,Wcant,A,B,Transaction)
              +Transaction(A).The_Task(B).Bij;
            for I in 1..Max_Transactions
            loop
               if I/=A then
                  MaxWik:=0.0;
                  for K in 1..Transaction(I).Ni
                  loop
                     if (Transaction(I).The_Task(K).Procij=Proc)
                       and(Transaction(I).The_Task(K).Prioij>=Prio)
                       and not((I=A)and(K=B))
                       and ((K=1)or else
                            (Transaction(I).The_Task(K-1).Procij/=Proc)
                            or else(Transaction(I).The_Task(K-1).Prioij<Prio)
                            or else Transaction(I).The_Task(K).Independent)
                     then
                        Wc_Ik:=Wik_LH(I,K,Wcant,A,B,Transaction);
                        if Wc_Ik>MaxWik then MaxWik:=Wc_Ik;end if;
                     end if;
                  end loop;
                  Wc:=Wc+MaxWik;
               end if;
            end loop;
            Done:=(Wc=Wcant);
            Wcant:=Wc;
            exit when Done;
         end loop;
         return Wc;
      end Busy_Period;

      ------------------
      -- Calculate_PL --
      ------------------

      function Calculate_PL
        (A:Transaction_ID_Type;
         B,C:Task_ID_Type;
         Transaction : in Linear_Transaction_System) return Long_Int
      is
         Prio:Priority;
      begin
         Prio:=Transaction(A).The_Task(B).Prioij;
         if In_MP(A,B,Prio,Transaction(A).The_Task(B).Procij,Transaction)
         then
            return Long_Int
              (Ceil0((Busy_Period(A,B,C,Transaction)-
                      FTijk(A,B,C,Transaction))/
                     Transaction(A).The_Task(C).Tij));
         elsif (C<B) and not Same_H(A,B,C,Prio,Transaction) then
            return Long_Int
              (-Time'Floor((Transaction(A).The_Task(C).Jij+
                            FTijk(A,C,C,Transaction))/Transaction(A).The_Task(C).Tij));
         else return 0;
         end if;
      exception
         when Constraint_Error => return Long_Int'Last;
      end Calculate_PL;

      P0,PL:Long_Int;
      R_Abc,W_Abc,Rmax:Time;
      Acum,Acumb:Time;
      Done:Boolean;
      Old:array(1..Max_Tasks_Per_Transaction) of Time;

      Transaction : Linear_Transaction_System;

   begin
      if Verbose then
         New_Line;Put_Line("Offset Based Analysis");
      end if;
      Translate_Linear_System(The_System,Transaction,Verbose);
      --show_linear_translation(Transaction);
      for A in 1..Max_Transactions
      loop
         Acum:=0.0;
         Acumb:=0.0;
         for B in 1..Transaction(A).Ni
         loop
            Acum:=Acum+Transaction(A).The_Task(B).Jinit;
            Transaction(A).The_Task(B).Jij:=Acum-Acumb;
            Acum:=Acum+Transaction(A).The_Task(B).Cijown;
            Transaction(A).The_Task(B).Rij:=Acum;
            Transaction(A).The_Task(B).Oij:=Acumb;
            if Transaction(A).The_Task(B).Oij >Transaction(A).The_Task(B).Tij
              and (Transaction(A).Kind_Of_Event=Sporadic or
               Transaction(A).Kind_Of_Event=Bursty) then
               for K in B..Transaction(A).Ni loop
                  Transaction(A).The_Task(B).Independent:=True;
               end loop;
            end if;
            Acumb:=Acumb+Transaction(A).The_Task(B).Cbijown;
            Transaction(A).The_Task(B).Rbij:=Acumb;
         end loop;
      end loop;

      loop
         Done:=True;
         for A in 1..Max_Transactions
         loop
            for B in 1..Transaction(A).Ni loop
               Old(B):=Transaction(A).The_Task(B).Rij;
            end loop;
            for B in 1..Transaction(A).Ni
            loop
               if Transaction(A).The_Task(B).Model /= Unbounded_Response
                 and Transaction(A).The_Task(B).Model /= Unbounded_Effects then
                  Rmax:=0.0;
                  for C in 1..Transaction(A).Ni
                  loop
                     if (Transaction(A).The_Task(C).Prioij>=
                         Transaction(A).The_Task(B).Prioij)
                       and(Transaction(A).The_Task(C).Procij=
                           Transaction(A).The_Task(B).Procij)
                       and Transaction(A).The_Task(B).Model /=
                       Unbounded_Effects
                       and
                       ((C=1) or else
                        (Transaction(A).The_Task(C-1).Procij/=
                         Transaction(A).The_Task(B).Procij)
                        or else (Transaction(A).The_Task(C-1).Prioij<
                                 Transaction(A).The_Task(B).Prioij)
                        or else Transaction(A).The_Task(B).Independent )
                     then
                        if Transaction(A).The_Task(B).Independent then
                           P0:=1;
                           PL:=Long_Int(Time'Ceiling((Busy_Period(A,B,C,Transaction)+
                                 Transaction(A).The_Task(B).Jij)/
                                 Transaction(A).The_Task(B).Tijown));
                        else
                           P0:=Long_Int
                             (-Time'Floor
                              ((Transaction(A).The_Task(B).Jij+
                                FTijk(A,B,C,Transaction))/Transaction(A).The_Task(B).Tijown)+
                              Time'(1.0));
                           PL:=Calculate_PL(A,B,C,Transaction);
                        end if;
                        if PL=Long_Int'Last then
                           for K in B..Transaction(A).Ni loop
                              Transaction(A).The_Task(K).Model:=
                                Unbounded_Effects;
                              Transaction(A).The_Task(K).Rij:=Large_Time;
                           end loop;
                        else
                           for P in P0..PL
                           loop
                              if Transaction(A).The_Task(B).Model/=
                                Unbounded_Effects then
                                 W_Abc:=W_DO_LH(A,B,C,P,Transaction);
                                 R_Abc:=W_Abc-FTijk(A,B,C,Transaction)-
                                   Time(P-1)*Transaction(A).The_Task(B).Tijown
                                   +Transaction(A).The_Task(B).Oij;
                                 if Transaction(A).The_Task(B).Independent then
                                    R_Abc:=Time(P-P0)*Transaction(A).The_task(B).Tijown+W_Abc+
                                      Transaction(A).The_Task(B).Oij+
                                      Transaction(A).The_Task(B).Jij;
                                 else
                                    R_Abc:=W_Abc-FTijk(A,B,C,Transaction)-
                                      Time(P-1)*Transaction(A).The_task(B).Tijown
                                      +Transaction(A).The_Task(B).Oij;
                                 end if;

                                 if R_Abc>Rmax then Rmax:=R_Abc;end if;
                                 if Rmax>Analysis_Bound
                                 then
                                    for K in B..Transaction(A).Ni loop
                                       Transaction(A).The_Task(K).Model:=
                                         Unbounded_Effects;
                                       Transaction(A).The_Task(K).Rij:=
                                         Large_Time;
                                    end loop;
                                 end if;
                              end if;
                           end loop;
                        end if;
                     end if;
                  end loop;
                  if Rmax>Transaction(A).The_Task(B).Rij then
                     Transaction(A).The_Task(B).Rij:=Rmax;
                  end if;

                  -------------------------------------------------
                  -- The analysis needs Oij+Jij to be in non
                  -- decreasing order for the tasks in a transaction
                  -- If we find that case (which case?) we estimate
                  -- for the second task an Rij equal to the last
                  -- task's Rij plus its own Cij
                  for L in B+1..Transaction(A).Ni
                  loop
                     if Transaction(A).The_Task(L).Rij<
                       Transaction(A).The_Task(L-1).Rij then
                        Transaction(A).The_Task(L).Rij:=
                          Transaction(A).The_Task(L-1).Rij+
                          Transaction(A).The_Task(L).Cij;
                     end if;
                  end loop;
                  -- As a result of applying Lemma 1, the estimation
                  -- for a task ij may be better than for the previous
                  -- task, because the Lemma may be applicable to that
                  -- task ij but not to the preceding one
                  -- In this case, we obtain a worst-case estimation for
                  -- task ij-1 making its response time equal to
                  -- Rij-1 = Rij-Cij}
                  for L in reverse 1..B-1
                  loop
                     if Transaction(A).The_Task(L).Rij>
                       Transaction(A).The_Task(L+1).Rij then
                        Transaction(A).The_Task(L).Rij:=
                          Transaction(A).The_Task(L+1).Rij-
                          Transaction(A).The_Task(L+1).Cij;
                     end if;
                  end loop;

                  -- Evaluate jitter
                  for L in 2..Transaction(A).Ni
                  loop
                     Transaction(A).The_Task(L).Jij:=
                       Transaction(A).The_Task(L-1).Rij-
                       Transaction(A).The_Task(L-1).Rbij;
                  end loop;
               end if;
            end loop;
            for B in 1..Transaction(A).Ni
            loop
               if Old(B)/=Transaction(A).The_Task(B).Rij then
                  Done:=False;
               end if;
            end loop;
         end loop;
         exit when Done;
      end loop;

      Translate_Linear_Analysis_Results(Transaction,The_System);
   end Offset_Based_Analysis;

end Mast.Linear_Analysis_Tools;
