-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with MAST.Systems,
  MAST.Transactions,
  Mast.Processing_Resources,
  Mast.Processing_Resources.FP,
  Mast.Scheduling_Parameters,
  Mast.Scheduling_Servers,
  Mast.Timers,
  Mast.Drivers;
use type Mast.Timers.System_Timer_Ref;

package body Mast.Max_Numbers is

   -----------------------------------------
   -- Calculate_Max_Transactions --
   -----------------------------------------

   function Calculate_Max_Transactions
     (The_System : Systems.System) return Natural
   is
      Max,Index :Natural;
      Proc_Ref : Processing_Resources.Processing_Resource_Ref;
      Proc_Iterator : Processing_Resources.Lists.Iteration_Object;
      Drv_Iterator : Processing_Resources.FP.Driver_Iteration_Object;
      Drv_Ref : Drivers.Driver_Ref;
   begin
      -- real transactions:
      Max:=Transactions.Lists.Size(The_System.Transactions);

      -- transactions caused by processing resource overheads
      Processing_Resources.Lists.Rewind
        (The_System.Processing_Resources,Proc_Iterator);
      for I in 1..Processing_Resources.Lists.Size
        (The_System.Processing_Resources)
      loop
         Processing_Resources.Lists.Get_Next_Item
           (Proc_Ref,The_System.Processing_Resources,Proc_Iterator);
         if Proc_Ref.all in
           Processing_Resources.FP.Fixed_Priority_Network'Class
         then
            -- add the transactions associated with the network drivers
            case Processing_Resources.FP.Transmission_Mode
              (Processing_Resources.FP.Fixed_Priority_Network
               (Proc_Ref.all)) is
               when Simplex | Half_duplex=>
                  Index:=1;
               when Full_Duplex =>
                  Index:=2;
            end case;
            Processing_Resources.FP.Rewind_Drivers
              (Processing_Resources.FP.Network(Proc_Ref.all),Drv_Iterator);
            for D in 1..Processing_Resources.FP.Num_Of_Drivers
              (Processing_Resources.FP.Network(Proc_Ref.all))
            loop
               Processing_Resources.FP.Get_Next_Driver
                 (Processing_Resources.FP.Network(Proc_Ref.all),
                  Drv_Ref,Drv_Iterator);
               if Drv_Ref.all in Drivers.Packet_Driver'Class then
                  if Drv_Ref.all in Drivers.Character_Packet_Driver'Class then
                     Max:=Max+2*Index;
                  else
                     Max:=Max+Index;
                  end if;
               end if;
            end loop;
         elsif Proc_Ref.all in
           Processing_Resources.FP.Fixed_Priority_Processor'Class
         then
            --add the transactions caused by Ticker system timers
            if Processing_Resources.FP.The_System_Timer
              (Processing_Resources.FP.Fixed_Priority_Processor
               (Proc_Ref.all))/=null and then
              Processing_Resources.FP.The_System_Timer
              (Processing_Resources.FP.Fixed_Priority_Processor
               (Proc_Ref.all)).all in Timers.Ticker'Class
            then
               Max:=Max+1;
            end if;
         else
            raise Incorrect_Object;
         end if;
      end loop;

      -- transactions caused by polling scheduler overheads

      declare
         Iterator : Scheduling_Servers.Lists.Iteration_Object;
         Srvr_Ref : Scheduling_Servers.Scheduling_Server_Ref;
      begin
         Scheduling_Servers.Lists.Rewind
           (The_System.Scheduling_Servers,Iterator);
         for I in 1..Scheduling_Servers.Lists.Size
           (The_System.Scheduling_Servers)
         loop
            Scheduling_Servers.Lists.Get_Next_Item
              (Srvr_Ref,The_System.Scheduling_Servers,Iterator);
            if Scheduling_Servers.Server_Sched_Parameters(Srvr_Ref.all).all in
              Scheduling_Parameters.Polling_Policy'Class
            then
               Max:=Max+1;
            end if;
         end loop;
      end;
      return Max;
   end Calculate_Max_Transactions;

   -----------------------------------------
   -- Calculate_Max_Tasks_Per_Transaction --
   -----------------------------------------

   function Calculate_Max_Tasks_Per_Transaction
     (The_System : Systems.System) return Natural
   is
      Max: Natural:=0;
      Current : Natural;
      Trans_Ref : Transactions.Transaction_Ref;
      Iterator : Transactions.Lists.Iteration_Object;
   begin
      Transactions.Lists.Rewind(The_System.Transactions,Iterator);
      for I in 1..Transactions.Lists.Size(The_System.Transactions)
      loop
         Transactions.Lists.Get_Next_Item
           (Trans_Ref,The_System.Transactions,Iterator);
         Current:=Transactions.Num_Of_Event_Handlers(Trans_Ref.all);
         if Current>Max then
            Max:=Current;
         end if;
      end loop;
      -- use double the number to take into account timer actions
      return 2*Max;
   end Calculate_Max_Tasks_Per_Transaction;

end Mast.Max_Numbers;
