-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2001                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@unican.es                     --
--          Jose Javier Gutierrez  gutierjj@unican.es                --
--          Jose Carlos Palencia   palencij@unican.es                --
--          Jose Maria Drake       drakej@unican.es                  --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with MAST.Transactions,Mast.Processing_Resources,
  Mast.Processing_Resources.FP,Mast.Timers,Mast.Drivers,
  Mast.Tool_Exceptions,Mast.Linear_Translation,
  Priority_Queues,Hash_Lists,Dynamic_Lists,List_Exceptions,
  Mast.Tools.Schedulability_Index;

with Ada.Float_Text_IO,Ada.Integer_Text_IO,Ada.Text_IO,
  Ada.Numerics.Discrete_Random,Ada.Numerics.Float_Random,
  Ada.Numerics.Generic_Elementary_Functions,
  Mast.Max_Numbers;

use type Mast.Timers.System_Timer_Ref;
use type Mast.Tools.Schedulability_Index.Index;
use Ada.Float_Text_IO,Ada.Integer_Text_IO,Ada.Text_IO;


package body Mast.Monoprocessor_Tools is


   Debug : constant Boolean := False;
   My_Verbose : constant Boolean := False;

   ------------------------
   -- Priority_Assignment --
   ------------------------

   procedure Priority_Assignment
     (The_System : in out MAST.Systems.System;
      The_Tool: in Mast.Tools.Worst_Case_Analysis_Tool;
      Verbose : in Boolean:=True)
   is

      type Processor_ID is new Natural;
      type Transaction_ID is new Natural;
      type Task_ID is new Natural;

      Max_Processors:constant Processor_ID:=Processor_ID
        (Processing_Resources.Lists.Size
         (The_System.Processing_Resources));
      Max_Transactions:constant Transaction_ID:=Transaction_ID(
        (Mast.Max_Numbers.Calculate_Max_Transactions(The_System)));
      Max_Tasks_Per_Transaction:constant Task_ID:=Task_ID
        (Mast.Max_Numbers.Calculate_Max_Tasks_Per_Transaction(The_System));

      subtype Processor_ID_Type is Processor_ID
        range 0..Max_Processors;
      subtype Transaction_ID_Type is Transaction_ID
        range 0..Max_Transactions;
      subtype Task_ID_Type is Task_ID
        range 0..Max_Tasks_Per_Transaction;

      package Translation is new Linear_Translation
        (Processor_ID_Type, Transaction_ID_Type, Task_ID_Type,
         Max_Processors, Max_Transactions, Max_Tasks_Per_Transaction);

      use Translation;

      My_System : Translation.Linear_Transaction_System;

      Queue_Length : Natural;
      List_Of_Min_VP : Virtual_Priority;
      List_Of_Min_P, List_Of_Max_P : Priority;
      Optimization_In_Processor : Boolean := True;

      type List_For_Tasks is array
        (Task_ID_Type range 1..Max_Tasks_Per_Transaction)
        of Translation.Priority_Assignment_Vars;

      type List_For_Transactions is array
        (Transaction_ID_Type range 1..Max_Transactions)
        of List_For_Tasks;

      Pavs : List_For_Transactions;

      Sched_Index : Mast.Tools.Schedulability_Index.Index;

      ---------------
      -- Save_Pavs --
      ---------------

      procedure Save_Pavs is
      begin
         for I in Transaction_ID_Type range 1..Max_Transactions
         loop
            for J in Task_ID_Type range 1..Max_Tasks_Per_Transaction
            loop
               Pavs(I)(J) := My_System(I).The_Task(J).Pav;
            end loop;
         end loop;
      end Save_Pavs;

      ------------------
      -- Restore_Pavs --
      ------------------

      procedure Restore_Pavs is
      begin
         for I in Transaction_ID_Type range 1..Max_Transactions
         loop
            for J in Task_ID_Type range 1..Max_Tasks_Per_Transaction
            loop
               My_System(I).The_Task(J).Pav := Pavs(I)(J);
            end loop;
         end loop;
      end Restore_Pavs;

      ------------------------------------
      -- Deadlines_Greater_Than_Periods --
      ------------------------------------

      function Deadlines_Greater_Than_Periods return Boolean is
      begin
         for I in Transaction_Id_Type range 1..Max_Transactions
         loop
            for J in 1..My_System(I).Ni
            loop
               if not My_System(I).The_Task(J).Pav.Hard_Prio then
                  if My_System(I).The_Task(J).Dij >
                    My_System(I).The_Task(J).Tij then
                     return True;
                  end if;
               end if;
            end loop;
         end loop;
         return False;
      end Deadlines_Greater_Than_Periods;

      ----------------------
      -- Priority_Mapping --
      ----------------------

      procedure Priority_Mapping is

         type Priority_Element is record
            Transaction : Transaction_Id_Type;
            Action : Task_Id_Type;
         end record;

         package VP_Lists is new Priority_Queues
           (Element => Priority_Element,
            Priority => Virtual_Priority,
            ">" => ">",
            "=" => "=");

         use VP_Lists;

         package P_Lists is new Priority_Queues
           (Element => Priority_Element,
            Priority => Priority,
            ">" => ">",
            "=" => "=");

         use P_Lists;

         Lists_For_VP : VP_Lists.Queue;
         Lists_For_P : P_Lists.Queue;

         Item1, Item2 : Priority_Element;
         Max_P, Min_P : Priority;
         Max_VP, Min_VP, VP : Virtual_Priority;

      begin

         for I in Transaction_ID_Type range 1..Max_Transactions
         loop
            for J in 1..My_System(I).Ni
            loop
               Item1 := (Transaction => I,
                         Action => J);
               if not My_System(I).The_Task(J).Pav.Hard_Prio then
                  if  My_System(I).The_Task(J).Pav.Preassigned then
                     P_Lists.Enqueue
                       (Item1,
                        My_System(I).The_Task(J).Pav.Preassigned_Prio,
                        Lists_For_P);
                  else
                     VP_Lists.Enqueue
                       (Item1,
                        My_System(I).The_Task(J).Pav.Virtual_Prio,
                        Lists_For_VP);
                  end if;
               end if;
            end loop;
         end loop;

         for I in Processor_ID_Type range 1..Max_Processors
         loop
            Max_P := List_Of_Max_P;
            Max_VP := Max_Virtual_Priority;
            if not P_Lists.Empty(Lists_For_P) then
               P_Lists.Dequeue(Item1,Min_P,Lists_For_P);
               Min_VP :=
                 My_System(Item1.Transaction).
                 The_Task(Item1.Action).Pav.Virtual_Prio;
            else
               Min_P := List_Of_Min_P;
               Min_VP := List_Of_Min_VP;
            end if;

            while not VP_Lists.Empty(Lists_For_VP)
            loop
               VP_Lists.Dequeue(Item2,VP,Lists_For_VP);
               if VP < Min_VP then
                  -- New stage
                  Max_P := Min_P;
                  Max_VP := Min_VP;
                  if not P_Lists.Empty(Lists_For_P) then
                     P_Lists.Dequeue(Item1,Min_P,Lists_For_P);
                     Min_VP :=
                       My_System(Item1.Transaction).
                       The_Task(Item1.Action).Pav.Virtual_Prio;
                  else
                     Min_P := List_Of_Min_P;
                     Min_VP := List_Of_Min_VP;
                  end if;
               end if;
               if Debug then
                  Put("   VP = "&Virtual_Priority'Image(VP));
                  Put("   Max_P = "&Priority'Image(Max_P));
                  Put("   Min_P = "&Priority'Image(Min_P));
                  Put("   Max_VP = "&Virtual_Priority'Image(Max_VP));
                  Put_Line("   Min_VP = "&Virtual_Priority'Image(Min_VP));
               end if;

               My_System(Item2.Transaction).The_Task(Item2.Action).Prioij :=
                 Priority
                 (Long_Float'Rounding
                  ((Long_Float(VP)-Long_Float(Min_VP))/
                   (Long_Float(Max_VP)-Long_Float(Min_VP)+1.0)*
                   (Long_Float(Max_P)-Long_Float(Min_P))+
                   Long_Float(Min_P)));

               if Debug then
                  Put_Line("   Priority Assigned = "&
                           Priority'Image
                           (My_System(Item2.Transaction).
                            The_Task(Item2.Action).Prioij));
               end if;

            end loop;
         end loop;

      end Priority_Mapping;

      -----------------------------------
      -- Deadline_Monotonic_Assignment --
      -----------------------------------

      procedure Deadline_Monotonic_Assignment is

         type Deadline_Element is record
            Transaction : Transaction_Id_Type;
            Action : Task_Id_Type;
         end record;

         package Deadlines_Lists is new Priority_Queues
           (Element => Deadline_Element,
            Priority => Time,
            ">" => "<",
            "=" => "=");

         use Deadlines_Lists;

         List_For_Processor : Deadlines_Lists.Queue;
         Item : Deadline_Element;
         Next_Priority : Virtual_Priority;
         Deadline : Time;

      begin

         -- Save Preassigned_Prio

         for I in Transaction_Id_Type range 1..Max_Transactions
         loop
            for J in 1..My_System(I).Ni
            loop
               My_System(I).The_Task(J).Pav.Preassigned_Prio :=
                 My_System(I).The_Task(J).Prioij;
            end loop;
         end loop;

         for I in Transaction_ID_Type range 1..Max_Transactions
         loop
            for J in 1..My_System(I).Ni
            loop
               if not My_System(I).The_Task(J).Pav.Hard_Prio then
                  Item := (Transaction => I,
                           Action => J);
                  Deadlines_Lists.Enqueue
                    (Item,
                     My_System(I).The_Task(J).Dij,
                     List_For_Processor);
               end if;
            end loop;
         end loop;

         for I in Processor_ID_Type range 1..Max_Processors
         loop
            declare
               J, K : Integer := 0;
            begin
               while not Deadlines_Lists.Empty(List_For_Processor)
               loop
                  J := J+1;
                  Deadlines_Lists.Dequeue
                    (Item,Deadline,List_For_Processor);
                  if J=1 then
                     Next_Priority := Max_Virtual_Priority;
                  else
                     Next_Priority := Virtual_Priority'Pred(Next_Priority);
                  end if;
                  if not My_System(Item.Transaction).
                    The_Task(Item.Action).Pav.Preassigned then
                     K := K+1;
                  end if;
                  My_System(Item.Transaction).
                    The_Task(Item.Action).Pav.Virtual_Prio := Next_Priority;
                  if My_System(Item.Transaction).The_Task(Item.Action).
                    Pav.Preassigned then
                     -- Check if the optimum priority assigment will be
                     -- performed for this processor
                     Optimization_In_Processor := False;
                  end if;
               end loop;
               Queue_Length := K;
               -- Store the number of task per processor for optimizing
               -- priority assignment
               List_Of_Min_VP := Next_Priority;
               -- Store the minimum virtual priority assigned for each
               -- processor
               List_Of_Min_P := Processing_Resources.Min_Priority
                 (My_System(Item.Transaction).The_Task(Item.Action).
                  Pav.P_R_Ref.all);
               -- Store the minimum priority for each processor
               List_Of_Max_P := Processing_Resources.Max_Priority
                 (My_System(Item.Transaction).The_Task(Item.Action).
                  Pav.P_R_Ref.all);
               -- Store the maximum priority for each processor
               if J > Integer
                 (Processing_Resources.Max_Priority
                  (My_System(Item.Transaction).The_Task(Item.Action).
                   Pav.P_R_Ref.all)-
                  Processing_Resources.Min_Priority
                  (My_System(Item.Transaction).The_Task(Item.Action).
                   Pav.P_R_Ref.all)) then
                  -- Check if the optimum priority assigment will be
                  -- performed for this processor
                  Optimization_In_Processor := False;
               end if;
            end;
         end loop;

         Priority_Mapping;

       end Deadline_Monotonic_Assignment;

      ---------------------------------
      -- Optimum_Priority_Assignment --
      ---------------------------------

      procedure Optimum_Priority_Assignment is

         type Task_Element is record
            Transaction : Transaction_Id_Type;
            Action : Task_Id_Type;
         end record;

         type Lists_Of_Task is
           array(Integer range 1..Integer(Max_Transactions))
           of Task_Element;
         List_Of_Task : Lists_Of_Task;
         Num_Of_Task : Integer := 0;
         Item : Task_Element;
         Prio : Priority;
         Failed : Boolean;


      begin

         -- Extract the tasks
         for I in Transaction_Id_Type range 1..Max_Transactions
         loop
            for J in 1..My_System(I).Ni
            loop
               if not My_System(I).The_Task(J).Pav.Hard_Prio then
                  Num_Of_Task := Num_Of_Task+1;
                  List_Of_Task(Num_Of_Task).Transaction := I;
                  List_Of_Task(Num_Of_Task).Action := J;
               end if;
            end loop;
         end loop;

         if Debug then
            for I in 1..Num_Of_Task
            loop
               Put("trans : "&Transaction_Id_Type'Image
                   (List_Of_Task(I).Transaction));
               Put("   deadline = "&time'Image
                   (My_System(List_Of_Task(I).Transaction).
                    The_Task(List_Of_Task(I).Action).Dij));
               Put_Line("   prioridad = "&Priority'Image
                        (My_System(List_Of_Task(I).Transaction).
                         The_Task(List_Of_Task(I).Action).Prioij));
            end loop;
            New_Line;
         end if;

         -- Order the task
         for I in 2..(Num_Of_Task)
         loop
           for J in reverse I..(Num_Of_Task)
           loop
              if My_System(List_Of_Task(J).Transaction).
                The_Task(List_Of_Task(J).Action).Prioij >
                My_System(List_Of_Task(J-1).Transaction).
                The_Task(List_Of_Task(J-1).Action).Prioij then
                 Item := List_Of_Task(J);
                 List_Of_Task(J) := List_Of_Task(J-1);
                 List_Of_Task(J-1) := Item;
              end if;
           end loop;
         end loop;

         if Debug then
            for I in 1..Num_Of_Task
            loop
               Put("trans : "&Transaction_Id_Type'Image
                   (List_Of_Task(I).Transaction));
               Put("   deadline = "&time'Image
                   (My_System(List_Of_Task(I).Transaction).
                    The_Task(List_Of_Task(I).Action).Dij));
               Put_Line("   prioridad = "&Priority'Image
                        (My_System(List_Of_Task(I).Transaction).
                         The_Task(List_Of_Task(I).Action).Prioij));
            end loop;
            New_Line;
         end if;

         -- Apply the priority assignment algorithm
         for I in reverse 2..Num_Of_Task
         loop
            Failed := True;
            for J in reverse 1..I
            loop
               -- Exchange tasks I and J and their priorities
               Item := List_Of_Task(I);
               List_Of_Task(I) := List_Of_Task(J);
               List_Of_Task(J) := Item;
               Prio := My_System(List_Of_Task(I).Transaction).
                 The_Task(List_Of_Task(I).Action).Prioij;
               My_System(List_Of_Task(I).Transaction).
                 The_Task(List_Of_Task(I).Action).Prioij :=
                 My_System(List_Of_Task(J).Transaction).
                 The_Task(List_Of_Task(J).Action).Prioij;
               My_System(List_Of_Task(J).Transaction).
                 The_Task(List_Of_Task(J).Action).Prioij := Prio;

               if Debug then
                  Put_Line("Change Priority.....");
                  for K in 1..Num_Of_Task
                  loop
                     Put("trans : "&Transaction_Id_Type'Image
                         (List_Of_Task(k).Transaction));
                     Put("   deadline = "&time'Image
                         (My_System(List_Of_Task(k).Transaction).
                          The_Task(List_Of_Task(k).Action).Dij));
                     Put_Line("   prioridad = "&Priority'Image
                              (My_System(List_Of_Task(k).Transaction).
                               The_Task(List_Of_Task(k).Action).Prioij));
                  end loop;
                  New_Line;
               end if;

               Translation.Translate_Priorities(My_System,The_System);
               Save_Pavs;

               -- Check if task I is schedulable

               Mast.Tools.Calculate_Ceilings_For_PCP_Resources
                 (The_System,My_Verbose);
               The_Tool(The_System,My_Verbose);
               Translation.Translate_Linear_System_With_Results
                 (The_System,My_System,My_Verbose);
               Restore_Pavs;
               if My_System(List_Of_Task(I).Transaction).
                 The_Task(List_Of_Task(I).Action).Dij >=
                 My_System(List_Of_Task(I).Transaction).
                 The_Task(List_Of_Task(I).Action).Rij then
                  Failed := False;
                  exit;
               else
                  -- Exchange tasks I and J and their old priorities
                  Item := List_Of_Task(I);
                  List_Of_Task(I) := List_Of_Task(J);
                  List_Of_Task(J) := Item;
                  Prio := My_System(List_Of_Task(I).Transaction).
                    The_Task(List_Of_Task(I).Action).Prioij;
                  My_System(List_Of_Task(I).Transaction).
                    The_Task(List_Of_Task(I).Action).Prioij :=
                    My_System(List_Of_Task(J).Transaction).
                    The_Task(List_Of_Task(J).Action).Prioij;
                  My_System(List_Of_Task(J).Transaction).
                    The_Task(List_Of_Task(J).Action).Prioij := Prio;
                  Translation.Translate_Priorities(My_System,The_System);
                  Save_Pavs;
               end if;
            end loop;
            exit when Failed;
         end loop;

         if Debug then
            for I in 1..Num_Of_Task
            loop
               Put("trans : "&Transaction_Id_Type'Image
                   (List_Of_Task(I).Transaction));
               Put("   deadline = "&time'Image
                   (My_System(List_Of_Task(I).Transaction).
                    The_Task(List_Of_Task(I).Action).Dij));
               Put_Line("   prioridad = "&Priority'Image
                        (My_System(List_Of_Task(I).Transaction).
                         The_Task(List_Of_Task(I).Action).Prioij));
            end loop;
            New_Line;
         end if;
      end Optimum_Priority_Assignment;


   begin
      Translation.Translate_Linear_System
        (The_System,My_System,My_Verbose);

      Deadline_Monotonic_Assignment;
      if Deadlines_Greater_Than_Periods and Optimization_In_Processor then
         Optimum_Priority_Assignment;
      end if;

      Translation.Translate_Priorities(My_System,The_System);

      if Verbose then
         Mast.Tools.Calculate_Ceilings_For_PCP_Resources
           (The_System,My_Verbose);
         The_Tool(The_System,My_Verbose);
         Translation.Translate_Linear_System_With_Results
           (The_System,My_System,My_Verbose);
         Mast.Tools.Schedulability_Index.Calculate_Schedulability_Index
           (The_System,Sched_Index,My_Verbose);
         Put("Monoprocessor Priority Assignment with "&
             "Schedulability Index : ");
         Mast.Tools.Schedulability_Index.Print(Sched_Index);
      end if;

   end Priority_Assignment;

   -----------------
   -- RM_Analysis --
   -----------------

   procedure RM_Analysis
     (The_System : in out MAST.Systems.System;
      Verbose : in Boolean:=True)
   is
   begin
      null;
   end RM_Analysis;

end Mast.Monoprocessor_Tools;














