-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with Ada.Text_IO,MAST.Shared_Resources,MAST.Scheduling_Parameters,
  Indexed_Lists, Named_Lists, Var_Strings;

package MAST.Operations is

   type Operation is abstract tagged private;

   procedure Init
     (Op : in out Operation;
      Name : Var_Strings.Var_String);
   function Name
     (Op : Operation) return Var_Strings.Var_String;

   function Worst_Case_Execution_Time
     (Op : Operation) return Normalized_Execution_Time is abstract;
   function Avg_Case_Execution_Time
     (Op : Operation) return Normalized_Execution_Time is abstract;
   function Best_Case_Execution_Time
     (Op : Operation) return Normalized_Execution_Time is abstract;

   procedure Scale
     (Op : in out Operation; Factor : Normalized_Execution_Time) is abstract;
   -- Scales all execution times multiplying them by factor

   procedure Set_New_Sched_Parameters
     (Op : in out Operation;
      Sched_Params :
        MAST.Scheduling_Parameters.Overridden_Sched_Parameters_Ref);
   function New_Sched_Parameters
     (Op : Operation)
      return MAST.Scheduling_Parameters.Overridden_Sched_Parameters_Ref;

   function Shared_Resources_Used
     (Op : Operation)
     return Boolean is abstract;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Operation;
      Indentation : Positive;
      Finalize    : Boolean:=False;
      Complete    : Boolean:=True);

   type Operation_Ref is access Operation'Class;

   function Clone
     (Op : Operation)
     return  Operation_Ref is abstract;

   procedure Adjust
     (Op : in out Operation;
      Res_List : in Shared_Resources.Lists.List) is abstract;
   -- To adjust internal pointers after clonning

   function Name (Op_Ref : Operation_Ref )
                  return Var_Strings.Var_String;

   package Lists is new Named_Lists
     (Element => Operation_Ref,
      Name    => Name);

   procedure Scale
     (The_List : in Lists.List; Factor : Normalized_Execution_Time);
   -- Scales all execution times multiplying them by factor

   function Clone
     (The_List : in Lists.List)
     return Lists.List;

   procedure Adjust
     (The_List : in Lists.List;
      Res_List : in Shared_Resources.Lists.List);
   -- To adjust internal pointers after clonning

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out Lists.List;
      Indentation : Positive);

   ------------------
   -- Simple Operation
   ------------------

   type Simple_Operation is new Operation with private;

   type Simple_Operation_Ref is access all Simple_Operation'Class;

   type Resource_Iteration_Object is private;

   procedure Set_Worst_Case_Execution_Time
     (Op : in out Simple_Operation;
      Worst_Case_Execution_Time :
        Normalized_Execution_Time);
   procedure Set_Avg_Case_Execution_Time
     (Op : in out Simple_Operation;
      Avg_Case_Execution_Time :
        Normalized_Execution_Time);
   procedure Set_Best_Case_Execution_Time
     (Op : in out Simple_Operation;
      Best_Case_Execution_Time :
        Normalized_Execution_Time);
   function Worst_Case_Execution_Time
     (Op : Simple_Operation) return Normalized_Execution_Time;
   function Avg_Case_Execution_Time
     (Op : Simple_Operation) return Normalized_Execution_Time;
   function Best_Case_Execution_Time
     (Op : Simple_Operation) return Normalized_Execution_Time;

   procedure Scale
     (Op : in out Simple_Operation; Factor : Normalized_Execution_Time);
   -- Scales all execution times multiplying them by factor

   procedure Add_Locked_Resource
     (Op : in out Simple_Operation;
      Resource : MAST.Shared_Resources.Shared_Resource_Ref);

   function Num_Of_Locked_Resources
     (Op : Simple_Operation )
      return Natural;

   procedure Rewind_Locked_Resources
     (Op : in Simple_Operation;
      Iterator : out Resource_Iteration_Object);

   procedure Get_Next_Locked_Resource
     (Op : in Simple_Operation;
      Resource : out MAST.Shared_Resources.Shared_Resource_Ref;
      Iterator : in out Resource_Iteration_Object);

   procedure Add_Unlocked_Resource
     (Op : in out Simple_Operation;
      Resource : MAST.Shared_Resources.Shared_Resource_Ref);

   function Num_Of_Unlocked_Resources
     (Op : Simple_Operation )
      return Natural;

   procedure Rewind_Unlocked_Resources
     (Op : in Simple_Operation;
      Iterator : out Resource_Iteration_Object);

   procedure Get_Next_Unlocked_Resource
     (Op : in Simple_Operation;
      Resource : out MAST.Shared_Resources.Shared_Resource_Ref;
      Iterator : in out Resource_Iteration_Object);

   procedure Add_Resource
     (Op : in out Simple_Operation;
      Resource : MAST.Shared_Resources.Shared_Resource_Ref);
   -- adds resource at the end of the list of locked resources
   -- and at the beginning of the list of unlocked resources

   function Shared_Resources_Used
     (Op : Simple_Operation)
     return Boolean;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Simple_Operation;
      Indentation : Positive;
      Finalize    : Boolean:=False;
      Complete    : Boolean:=True);

   function Clone
     (Op : Simple_Operation)
     return  Operation_Ref;
   -- Lists of locked and unlocked resources are not cloned
   -- They need to be rearranged later

   procedure Adjust
     (Op : in out Simple_Operation;
      Res_List : in Shared_Resources.Lists.List);
   -- To adjust internal pointers after clonning

   -------------------
   -- Composite Operation
   -------------------

   type Composite_Operation is new Operation with private;

   type Operation_Iteration_Object is private;

   function Worst_Case_Execution_Time
     (Op : Composite_Operation) return Normalized_Execution_Time;
   function Avg_Case_Execution_Time
     (Op : Composite_Operation) return Normalized_Execution_Time;
   function Best_Case_Execution_Time
     (Op : Composite_Operation) return Normalized_Execution_Time;

   procedure Scale
     (Op : in out Composite_Operation; Factor : Normalized_Execution_Time);
   -- Scales all execution times multiplying them by factor

   procedure Add_Operation
     (Comp_Op : in out Composite_Operation;
      Op : Operation_Ref);

   function Num_Of_Operations
     (Comp_Op : Composite_Operation)
      return Natural;

   procedure Rewind_Operations
     (Comp_Op : in Composite_Operation;
      Iterator : out Operation_Iteration_Object);

   procedure Get_Next_Operation
     (Comp_Op : in Composite_Operation;
      Op : out Operation_Ref;
      Iterator : in out Operation_Iteration_Object);

   function Shared_Resources_Used
     (Op : Composite_Operation)
     return Boolean;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Composite_Operation;
      Indentation : Positive;
      Finalize    : Boolean:=False;
      Complete    : Boolean:=True);

   function Clone
     (Op : Composite_Operation)
     return  Operation_Ref;

   procedure Adjust
     (Op : in out Composite_Operation;
      Res_List : in Shared_Resources.Lists.List);
   -- To adjust internal pointers after clonning

   ----------------------
   -- Enclosing Operation
   ----------------------

   type Enclosing_Operation is new Composite_Operation with private;

   procedure Set_Worst_Case_Execution_Time
     (Op : in out Enclosing_Operation;
      Worst_Case_Execution_Time : Normalized_Execution_Time);
   procedure Set_Avg_Case_Execution_Time
     (Op : in out Enclosing_Operation;
      Avg_Case_Execution_Time : Normalized_Execution_Time);
   procedure Set_Best_Case_Execution_Time
     (Op : in out Enclosing_Operation;
      Best_Case_Execution_Time : Normalized_Execution_Time);

   function Worst_Case_Execution_Time
     (Op : Enclosing_Operation) return Normalized_Execution_Time;
   function Avg_Case_Execution_Time
     (Op : Enclosing_Operation) return Normalized_Execution_Time;
   function Best_Case_Execution_Time
     (Op : Enclosing_Operation) return Normalized_Execution_Time;

   procedure Scale
     (Op : in out Enclosing_Operation; Factor : Normalized_Execution_Time);
   -- Scales all execution times multiplying them by factor

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Enclosing_Operation;
      Indentation : Positive;
      Finalize    : Boolean:=False;
      Complete    : Boolean:=True);

   function Clone
     (Op : Enclosing_Operation)
     return  Operation_Ref;

   -- Adjust is inherited from composite

private

   package Shared_Resources_List is new Indexed_Lists
     (Element => MAST.Shared_Resources.Shared_Resource_Ref,
      "=" => MAST.Shared_Resources."=");

   type Resource_Iteration_Object is new
     Shared_Resources_List.Index;

   type Operation is abstract tagged record
      Name : Var_Strings.Var_String;
      New_Sched_Parameters :
        MAST.Scheduling_Parameters.Overridden_Sched_Parameters_Ref;
   end record;

   type Simple_Operation is new Operation with record
      Worst_Case_Execution_Time,
      Avg_Case_Execution_Time : Normalized_Execution_Time
      :=Large_Execution_Time;
      Best_Case_Execution_Time  : Normalized_Execution_Time
        :=0.0;
      Scale_Factor : Normalized_Execution_Time:=1.0;
      Locked_Shared_Resources,
      Unlocked_Shared_Resources : Shared_Resources_List.List;
   end record;

   package Op_Lists is new Indexed_Lists
     (Element => Operation_Ref,
      "="     => "=");

   type Operation_Iteration_Object is new Op_Lists.Index;

   type Composite_Operation is new Operation with record
      Op_List : Op_Lists.List;
   end record;

   type Enclosing_Operation is new Composite_Operation with record
      Worst_Case_Execution_Time,
      Avg_Case_Execution_Time : Normalized_Execution_Time
      :=Large_Execution_Time;
      Best_Case_Execution_Time  : Normalized_Execution_Time
        :=0.0;
      Scale_Factor : Normalized_Execution_Time:=1.0;
   end record;


end MAST.Operations;
