-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with List_Exceptions, MAST.Graphs.Links, MAST.IO,
  Ada.Strings,Ada.Strings.Fixed,Ada.Text_IO,Ada.Float_Text_IO,
  Var_Strings;
use Ada.Strings,Ada.Strings.Fixed,Ada.Text_IO,Ada.Float_Text_IO;
use type MAST.Graphs.Link_Ref;

package body MAST.Results is

   Names_Length : constant Positive := 29;

   -----------------------
   -- Link_Of           --
   -----------------------

   function Link_Of
     (The_Result : Result)
      return MAST.Graphs.Link_Ref
   is
   begin
      return The_Result.The_Link;
   end Link_Of;

   ---------------------------
   -- Set_Link              --
   ---------------------------

   procedure Set_Link
     (The_Result : in out Result;
      The_Link : MAST.Graphs.Link_Ref)
   is
   begin
      The_Result.The_Link := The_Link;
   end Set_Link;

   --------------------------
   -- Create_Timing_Result --
   --------------------------

   function Create_Timing_Result
     (The_Link_Ref : MAST.Graphs.Link_Ref)
      return Result_Ref
   is
      Res_Ref : Result_Ref;
   begin
      Res_Ref  := new Timing_Result;
      MAST.Graphs.Links.Set_Link_Time_Results
        (Mast.Graphs.Links.Regular_Link(The_Link_Ref.all),
         Res_Ref);
      Set_Link(Res_Ref.all,The_Link_Ref);
      return Res_Ref;
   end Create_Timing_Result;

   ----------------------------------------------------------------
   --               Worst Case Analysis Parameters               --
   ----------------------------------------------------------------

   -------------------------------
   -- Best_Global_Response_Time --
   -------------------------------

   function Best_Global_Response_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
   begin
      return Time_Result_Lists.Item
        (The_Event,The_Result.Best_Global_Response_Times);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Best_Global_Response_Time;

   -------------------------------
   -- Best_Global_Response_Time --
   -------------------------------

   function Best_Global_Response_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return Time_Result_Lists.First_Item
        (The_Result.Best_Global_Response_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Best_Global_Response_Time;

   -------------------
   -- Blocking_Time --
   -------------------

   function Blocking_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return The_Result.Blocking_Time;
   end Blocking_Time;

   ----------------------------------------
   -- Get_Next_Global_Timing_Parameters --
   ----------------------------------------

   procedure Get_Next_Global_Timing_Parameters
     (The_Result : in Timing_Result;
      The_Event : out MAST.Events.Event_Ref;
      Timing_Params : out Timing_Parameters;
      Iterator : in out Time_Iteration_Object)
   is
   begin
      Time_Result_Lists.Get_Next_Item
        (The_Event,Timing_Params.Worst_Response_Time,
         The_Result.Worst_Global_Response_Times,Iterator.Worst);
      Time_Result_Lists.Get_Next_Item
        (The_Event,Timing_Params.Best_Response_Time,
         The_Result.Best_Global_Response_Times,Iterator.Best);
      Time_Result_Lists.Get_Next_Item
        (The_Event,Timing_Params.Jitter,
         The_Result.Jitters,Iterator.Jitters);
   end Get_Next_Global_Timing_Parameters;

   ------------
   -- Jitter --
   ------------

   function Jitter
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
   begin
      return Time_Result_Lists.Item
        (The_Event,The_Result.Jitters);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Jitter;

   ------------
   -- Jitter --
   ------------

   function Jitter
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return Time_Result_Lists.First_Item
        (The_Result.Jitters);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Jitter;

   ---------------------------------------
   -- Num_Of_Global_Timing_Parameters --
   ---------------------------------------

   function Num_Of_Global_Timing_Parameters
     (The_Result : Timing_Result)
      return Natural
   is
   begin
      return Time_Result_Lists.Size
        (The_Result.Worst_Global_Response_Times);
   end Num_Of_Global_Timing_Parameters;

   ---------------------------------------
   -- Rewind_Global_Timing_Parameters --
   ---------------------------------------

   procedure Rewind_Global_Timing_Parameters
     (The_Result : in Timing_Result;
      Iterator : out Time_Iteration_Object)
   is
   begin
      Time_Result_Lists.Rewind
        (The_Result.Worst_Global_Response_Times,Iterator.Worst);
      Time_Result_Lists.Rewind
        (The_Result.Best_Global_Response_Times,Iterator.Best);
      Time_Result_Lists.Rewind
        (The_Result.Jitters,Iterator.Jitters);
   end Rewind_Global_Timing_Parameters;

   -----------------------------------
   -- Set_Best_Global_Response_Time --
   -----------------------------------

   procedure Set_Best_Global_Response_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Best_Global_Response_Time : Time)
   is
   begin
      The_Result.Has_Best_Global_Response_Times:=True;
      Time_Result_Lists.Add_Or_Update
        (The_Event,Best_Global_Response_Time,
         The_Result.Best_Global_Response_Times);
   end Set_Best_Global_Response_Time;

   -----------------------------------
   -- Set_Best_Global_Response_Time --
   -----------------------------------

   procedure Set_Best_Global_Response_Time
     (The_Result : in out Timing_Result;
      Best_Global_Response_Time : Time)
   is
   begin
      The_Result.Has_Best_Global_Response_Times:=True;
      Time_Result_Lists.Update_First
        (Best_Global_Response_Time,
         The_Result.Best_Global_Response_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Set_Best_Global_Response_Time;

   -----------------------
   -- Set_Blocking_Time --
   -----------------------

   procedure Set_Blocking_Time
     (The_Result : in out Timing_Result;
      Blocking_Time : Time)
   is
   begin
      The_Result.Has_Blocking_Time:=True;
      The_Result.Blocking_Time := Blocking_Time;
   end Set_Blocking_Time;

   ----------------
   -- Set_Jitter --
   ----------------

   procedure Set_Jitter
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Jitter : Time)
   is
   begin
      The_Result.Has_Worst_Global_Response_Times:=True;
      Time_Result_Lists.Add_Or_Update
        (The_Event,Jitter,
         The_Result.Jitters);
   end Set_Jitter;

   ----------------
   -- Set_Jitter --
   ----------------

   procedure Set_Jitter
     (The_Result : in out Timing_Result;
      Jitter : Time)
   is
   begin
      The_Result.Has_Worst_Global_Response_Times:=True;
      Time_Result_Lists.Update_First
        (Jitter,
         The_Result.Jitters);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Set_Jitter;

   ------------------------------------
   -- Set_Worst_Global_Response_Time --
   ------------------------------------

   procedure Set_Worst_Global_Response_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Worst_Global_Response_Time : Time)
   is
   begin
      The_Result.Has_Worst_Global_Response_Times:=True;
      Time_Result_Lists.Add_Or_Update
        (The_Event,Worst_Global_Response_Time,
         The_Result.Worst_Global_Response_Times);
   end Set_Worst_Global_Response_Time;

   ------------------------------------
   -- Set_Worst_Global_Response_Time --
   ------------------------------------

   procedure Set_Worst_Global_Response_Time
     (The_Result : in out Timing_Result;
      Worst_Global_Response_Time : Time)
   is
   begin
      The_Result.Has_Worst_Global_Response_Times:=True;
      Time_Result_Lists.Update_First
        (Worst_Global_Response_Time,
         The_Result.Worst_Global_Response_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Set_Worst_Global_Response_Time;

   -----------------------------------
   -- Set_Worst_Local_Response_Time --
   -----------------------------------

   procedure Set_Worst_Local_Response_Time
     (The_Result : in out Timing_Result;
      Worst_Local_Response_Time : Time)
   is
   begin
      The_Result.Has_Worst_Local_Response_Time:=True;
      The_Result.Worst_Local_Response_Time :=
        Worst_Local_Response_Time;
   end Set_Worst_Local_Response_Time;

   --------------------------------
   -- Worst_Global_Response_Time --
   --------------------------------

   function Worst_Global_Response_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
   begin
      return Time_Result_Lists.Item
        (The_Event,The_Result.Worst_Global_Response_Times);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Worst_Global_Response_Time;

   --------------------------------
   -- Worst_Global_Response_Time --
   --------------------------------

   function Worst_Global_Response_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return Time_Result_Lists.First_Item
        (The_Result.Worst_Global_Response_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Worst_Global_Response_Time;

   -------------------------------
   -- Worst_Local_Response_Time --
   -------------------------------

   function Worst_Local_Response_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return Time_Result_Lists.First_Item
        (The_Result.Worst_Global_Response_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Worst_Local_Response_Time;


   ---------------------------------------------------------------
   --                  Simulation Parameters                    --
   ---------------------------------------------------------------

   --------------------------------
   -- Avg_Global_Simulation_Time --
   --------------------------------

   function Avg_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
      Params : Simulation_Parameters;
   begin
      Params:=Sim_Params_Lists.Item
        (The_Event,The_Result.Global_Simulation_Times);
      return Params.Sum_Of_Simulation_Times/
        Time(Params.Num_Of_Simulation_Times);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Avg_Global_Simulation_Time;

   ------------------------------
   -- Avg_Global_Simulation_Time --
   ------------------------------

   function Avg_Global_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
      Params : Simulation_Parameters;
   begin
      Params:=Sim_Params_Lists.First_Item
        (The_Result.Global_Simulation_Times);
      return Params.Sum_Of_Simulation_Times/
        Time(Params.Num_Of_Simulation_Times);
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Avg_Global_Simulation_Time;

   -----------------------------
   -- Avg_Local_Simulation_Time --
   -----------------------------

   function Avg_Local_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return The_Result.Local_Simulation_Time.
        Sum_Of_Simulation_Times/
        Time(The_Result.Local_Simulation_Time.
             Num_Of_Simulation_Times);
   end Avg_Local_Simulation_Time;

   -------------------------------
   -- Best_Global_Simulation_Time --
   -------------------------------

   function Best_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
   begin
      return Sim_Params_Lists.Item
        (The_Event,The_Result.Global_Simulation_Times).
        Best_Simulation_Time;
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Best_Global_Simulation_Time;

   -------------------------------
   -- Best_Global_Simulation_Time --
   -------------------------------

   function Best_Global_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return The_Result.Local_Simulation_Time.Best_Simulation_Time;
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Best_Global_Simulation_Time;

   ------------------------------
   -- Best_Local_Simulation_Time --
   ------------------------------

   function Best_Local_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return The_Result.Local_Simulation_Time.Best_Simulation_Time;
   end Best_Local_Simulation_Time;

   ---------------------------------------
   -- Get_Next_Global_Simulation_Times --
   ---------------------------------------

   procedure Get_Next_Global_Simulation_Times
     (The_Result : in Timing_Result;
      The_Event : out MAST.Events.Event_Ref;
      Global_Simulation_Params : out Simulation_Parameters;
      Iterator : in out Sim_Iteration_Object)
   is
   begin
      Sim_Params_Lists.Get_Next_Item
        (The_Event,Global_Simulation_Params,
         The_Result.Global_Simulation_Times,
         Sim_Params_Lists.Iteration_Object(Iterator));
   end Get_Next_Global_Simulation_Times;

   -------------------------------
   -- Get_Next_Local_Miss_Ratio --
   -------------------------------

   procedure Get_Next_Local_Miss_Ratio
     (The_Result : in Timing_Result;
      The_Deadline : out Time;
      The_Ratio : out Percentage;
      Iterator : in out Local_Ratio_Iteration_Object)
   is
      The_Data : Miss_Ratio_Data;
   begin
      Miss_Ratio_Lists.Get_Next_Item
        (The_Deadline, The_Data,
         The_Result.Local_Miss_Ratios,
         Miss_Ratio_Lists.Iteration_Object(Iterator));
      The_Ratio:=Float(The_Data.Num_Missed)/Float(The_Data.Total);
   end Get_Next_Local_Miss_Ratio;

   -------------------------------
   -- Get_Next_Global_Miss_Ratio --
   -------------------------------

   procedure Get_Next_Global_Miss_Ratio
     (The_Result : in Timing_Result;
      The_Deadline : out Time;
      The_Event : out MAST.Events.Event_Ref;
      The_Ratio : out Percentage;
      Iterator : in out Global_Ratio_Iteration_Object)
   is
      The_Data : Miss_Ratio_Data;
   begin
      if Iterator.Current_List=null then
         raise List_Exceptions.No_More_Items;
      else
         begin
            Miss_Ratio_Lists.Get_Next_Item
              (The_Deadline, The_Data,
               Iterator.Current_List.all,Iterator.Miss_List_Iterator);
            The_Ratio:=Float(The_Data.Num_Missed)/Float(The_Data.Total);
            The_Event:=Iterator.Current_Event;
         exception
            when List_Exceptions.No_More_Items =>
               List_Of_Miss_Lists.Get_Next_Item
                 (Iterator.Current_Event,Iterator.Current_List,
                  The_Result.Global_Miss_Ratios,
                  Iterator.List_Of_Lists_Iterator);
               Miss_Ratio_Lists.Rewind
                 (Iterator.Current_List.all,Iterator.Miss_List_Iterator);
               Get_Next_Global_Miss_Ratio
                 (The_Result,The_Deadline,The_Event,The_Ratio,Iterator);
         end;
      end if;
   end Get_Next_Global_Miss_Ratio;

   ----------------------
   -- Local_Miss_Ratio --
   ----------------------

   function Local_Miss_Ratio
     (The_Result : Timing_Result;
      The_Deadline : Time)
      return Percentage
   is
      The_Data : Miss_Ratio_Data;
   begin
      The_Data:= Miss_Ratio_Lists.Item
        (The_Deadline,The_Result.Local_Miss_Ratios);
      return Float(The_Data.Num_Missed)/Float(The_Data.Total);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Deadline;
   end Local_Miss_Ratio;

   -----------------------
   -- Global_Miss_Ratio --
   -----------------------

   function Global_Miss_Ratio
     (The_Result : Timing_Result;
      The_Deadline : Time;
      The_Event : MAST.Events.Event_Ref)
      return Percentage
   is
      The_List : List_Ref;
      The_Data : Miss_Ratio_Data;
   begin
      begin
         The_List:=List_Of_Miss_Lists.Item
           (The_Event,The_Result.Global_Miss_Ratios);
      exception
         when List_Exceptions.Invalid_Index =>
            raise No_Results_For_Event;
      end;
      The_Data:=Miss_Ratio_Lists.Item(The_Deadline,The_List.all);
      return Float(The_Data.Num_Missed)/Float(The_Data.Total);
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Deadline;
   end Global_Miss_Ratio;

   --------------------------------------
   -- Num_Of_Global_Simulation_Times --
   --------------------------------------

   function Num_Of_Global_Simulation_Times
     (The_Result : Timing_Result)
      return Natural
   is
   begin
      return Sim_Params_Lists.Size
        (The_Result.Global_Simulation_Times);
   end Num_Of_Global_Simulation_Times;

   ------------------------------
   -- Num_Of_Local_Miss_Ratios --
   ------------------------------

   function Num_Of_Local_Miss_Ratios
     (The_Result : Timing_Result)
      return Natural
   is
   begin
      return Miss_Ratio_Lists.Size
        (The_Result.Local_Miss_Ratios);
   end Num_Of_Local_Miss_Ratios;

   -------------------------------
   -- Num_Of_Global_Miss_Ratios --
   -------------------------------

   function Num_Of_Global_Miss_Ratios
     (The_Result : Timing_Result)
      return Natural
   is
      Count : Natural:=0;
      The_List : List_Ref;
      The_Event : MAST.Events.Event_Ref;
      Iterator : List_Of_Miss_Lists.Iteration_Object;
   begin
      List_Of_Miss_Lists.Rewind(The_Result.Global_Miss_Ratios,Iterator);
      for I in 1..List_Of_Miss_Lists.Size(The_Result.Global_Miss_Ratios)
      loop
         List_Of_Miss_Lists.Get_Next_Item
           (The_Event,The_List,The_Result.Global_Miss_Ratios,Iterator);
         Count:=Count+Miss_Ratio_Lists.Size(The_List.all);
      end loop;
      return Count;
   end Num_Of_Global_Miss_Ratios;

   ----------------------------
   -- Reset_Local_Miss_Ratio --
   ----------------------------

   procedure Reset_Local_Miss_Ratio
     (The_Result : in out Timing_Result;
      The_Deadline : Time)
   is
      Ratio_Data : Miss_Ratio_Data;
   begin
      The_Result.Has_Simulation_Times:=True;
      Ratio_Data:=Miss_Ratio_Lists.Item
        (The_Deadline,The_Result.LOcal_Miss_Ratios);
      Miss_Ratio_Lists.Update
        (The_Deadline,Miss_Ratio_Data'(0,0),
         The_Result.Local_Miss_Ratios);
   exception
      when List_Exceptions.Invalid_Index =>
         Miss_Ratio_Lists.Add
           (The_Deadline,Miss_Ratio_Data'(0,0),
            The_Result.Local_Miss_Ratios);
   end Reset_Local_Miss_Ratio;

   ------------------------------
   -- Reset_Global_Miss_Ratios --
   ------------------------------

   procedure Reset_Global_Miss_Ratio
     (The_Result : in out Timing_Result;
      The_Deadline : Time;
      The_Event : MAST.Events.Event_Ref)
   is
      Ratio_Data : Miss_Ratio_Data;
      Ratio_List : List_Ref;
   begin
      The_Result.Has_Simulation_Times:=True;
      Ratio_List:=List_Of_Miss_Lists.Item
        (The_Event,The_Result.Global_Miss_Ratios);
      begin
         Ratio_Data:=Miss_Ratio_Lists.Item
           (The_Deadline,Ratio_List.all);
         Miss_Ratio_Lists.Update
           (The_Deadline,Miss_Ratio_Data'(0,0),Ratio_List.all);
      exception
         when List_Exceptions.Invalid_Index =>
            Miss_Ratio_Lists.Add
              (The_Deadline,Miss_Ratio_Data'(0,0),Ratio_List.all);
      end;
   exception
      when List_Exceptions.Invalid_Index =>
         Ratio_List:=new Miss_Ratio_Lists.List;
         Miss_Ratio_Lists.Add
           (The_Deadline,Miss_Ratio_Data'(0,0),Ratio_List.all);
         List_Of_Miss_Lists.Add
           (The_Event,Ratio_List,The_Result.Global_Miss_Ratios);
   end Reset_Global_Miss_Ratio;

   --------------------------------------
   -- Rewind_Global_Simulation_Times --
   --------------------------------------

   procedure Rewind_Global_Simulation_Times
     (The_Result : in Timing_Result;
      Iterator : out Sim_Iteration_Object)
   is
   begin
      Sim_Params_Lists.Rewind
        (The_Result.Global_Simulation_Times,
         Sim_Params_Lists.Iteration_Object(Iterator));
   end Rewind_Global_Simulation_Times;

   ------------------------------
   -- Rewind_Local_Miss_Ratios --
   ------------------------------

   procedure Rewind_Local_Miss_Ratios
     (The_Result : in Timing_Result;
      Iterator : out Local_Ratio_Iteration_Object)
   is
   begin
      Miss_Ratio_Lists.Rewind
        (The_Result.Local_Miss_Ratios,
         Miss_Ratio_Lists.Iteration_Object(Iterator));
   end Rewind_Local_Miss_Ratios;

   ------------------------------
   -- Rewind_Global_Miss_Ratios --
   ------------------------------

   procedure Rewind_Global_Miss_Ratios
     (The_Result : in Timing_Result;
      Iterator : out Global_Ratio_Iteration_Object)
   is
   begin
      List_Of_Miss_Lists.Rewind
        (The_Result.Global_Miss_Ratios,Iterator.List_Of_Lists_Iterator);
      List_Of_Miss_Lists.Get_Next_Item
        (Iterator.Current_Event, Iterator.Current_List,
         The_Result.Global_Miss_Ratios,Iterator.List_Of_Lists_Iterator);
      Miss_Ratio_Lists.Rewind
        (Iterator.Current_List.all,Iterator.Miss_List_Iterator);
   exception
      when List_Exceptions.No_More_Items =>
         Iterator.Current_List:=null;
         Iterator.Current_Event:=null;
   end Rewind_Global_Miss_Ratios;

   ----------------------------------
   -- Update Simulation Parameters --
   ----------------------------------

   procedure Update
     (The_Time : Time; Params : in out Simulation_Parameters)
   is
   begin
      if The_Time>Params.Worst_Simulation_Time then
         Params.Worst_Simulation_Time:=The_Time;
      end if;
      if The_Time<Params.Best_Simulation_Time then
         Params.Best_Simulation_Time:=The_Time;
      end if;
      Params.Sum_Of_Simulation_Times:=
        Params.Sum_Of_Simulation_Times+The_Time;
      Params.Num_Of_Simulation_Times:=
        Params.Num_Of_Simulation_Times+1;
   end Update;


   ----------------------------------
   -- Update Miss Ratio Data       --
   ----------------------------------

   procedure Update
     (The_Time : Time;
      The_Deadline : Time;
      Data : in out Miss_Ratio_Data)
   is
   begin
      if The_Time>The_Deadline then
         Data.Num_Missed:=Data.Num_Missed+1;
      end if;
      Data.Total:=Data.Total+1;
   end Update;

   -----------------------------------
   -- Record Local simulation times --
   -----------------------------------

   procedure Record_Local_Simulation_Time
     (The_Result : in out Timing_Result;
      Local_Simulation_Time : Time)
   is
      The_Deadline : Time;
      The_Ratio_Data : Miss_Ratio_Data;
      Iterator : Miss_Ratio_Lists.Iteration_Object;
   begin
      The_Result.Has_Simulation_Times:=True;
      Update (Local_Simulation_Time,
              The_Result.Local_Simulation_Time);
      Miss_Ratio_Lists.Rewind(The_Result.Local_Miss_Ratios,Iterator);
      for I in 1..Miss_Ratio_Lists.Size
        (The_Result.Local_Miss_Ratios)
      loop
         Miss_Ratio_Lists.Get_Next_Item
           (The_Deadline,The_Ratio_Data,
            The_Result.Local_Miss_Ratios,Iterator);
         Update(Local_Simulation_Time,The_Deadline,The_Ratio_Data);
         Miss_Ratio_Lists.Update
           (The_Deadline,The_Ratio_Data,
            The_Result.Local_Miss_Ratios);
      end loop;
   end Record_Local_Simulation_Time;

   ----------------------------------
   -- Record_Global_Simulation_Time --
   ----------------------------------

   procedure Record_Global_Simulation_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Global_Simulation_Time : Time)
   is
      The_Miss_List : List_Ref;
      Params : Simulation_Parameters;
      The_Deadline : Time;
      The_Ratio_Data : Miss_Ratio_Data;
      Iterator : Miss_Ratio_Lists.Iteration_Object;
   begin
      The_Result.Has_Simulation_Times:=True;
      -- Update or create simulation parameters
      begin
         Params:=Sim_Params_Lists.Item
           (The_Event,
            The_Result.Global_Simulation_Times);
         Update (Global_Simulation_Time,Params);
         Sim_Params_Lists.Update
           (The_Event,Params,
            The_Result.Global_Simulation_Times);
      exception
         when List_Exceptions.Invalid_Index =>
            Params:=
              (Worst_Simulation_Time  =>Global_Simulation_Time,
               Best_Simulation_Time   =>Global_Simulation_Time,
               Sum_Of_Simulation_Times=>Global_Simulation_Time,
               Num_Of_Simulation_Times=>1);
            Sim_Params_Lists.Add
              (The_Event,Params,
               The_Result.Global_Simulation_Times);
      end;
      -- update (but not create) global miss ratios
      The_Miss_List:=List_Of_Miss_Lists.Item
        (The_Event,The_Result.Global_Miss_Ratios);
      Miss_Ratio_Lists.Rewind(The_Miss_List.all,Iterator);
      for I in 1..Miss_Ratio_Lists.Size(The_Miss_List.all) loop
         Miss_Ratio_Lists.Get_Next_Item
           (The_Deadline,The_Ratio_Data,The_Miss_List.all,Iterator);
         Update(Global_Simulation_Time,The_Deadline,The_Ratio_Data);
         Miss_Ratio_Lists.Update
           (The_Deadline,The_Ratio_Data,The_Miss_List.all);
      end loop;
   end Record_Global_Simulation_Time;

   ----------------------------------
   -- Record_Global_Simulation_Time --
   ----------------------------------

   procedure Record_Global_Simulation_Time
     (The_Result : in out Timing_Result;
      Global_Simulation_Time : Time)
   is
      The_Miss_List : List_Ref;
      Params : Simulation_Parameters;
      The_Deadline : Time;
      The_Ratio_Data : Miss_Ratio_Data;
      Iterator : Miss_Ratio_Lists.Iteration_Object;
   begin
      The_Result.Has_Simulation_Times:=True;
      -- Update (but not create) simulation parameters
      Params:=Sim_Params_Lists.First_Item
        (The_Result.Global_Simulation_Times);
      Update (Global_Simulation_Time,Params);
      Sim_Params_Lists.Update_First
        (Params,The_Result.Global_Simulation_Times);
      -- update (but not create) global miss ratios
      The_Miss_List:=List_Of_Miss_Lists.First_Item
        (The_Result.Global_Miss_Ratios);
      Miss_Ratio_Lists.Rewind(The_Miss_List.all,Iterator);
      for I in 1..Miss_Ratio_Lists.Size(The_Miss_List.all) loop
         Miss_Ratio_Lists.Get_Next_Item
           (The_Deadline,The_Ratio_Data,The_Miss_List.all,Iterator);
         Update(Global_Simulation_Time,The_Deadline,The_Ratio_Data);
         Miss_Ratio_Lists.Update
           (The_Deadline,The_Ratio_Data,The_Miss_List.all);
      end loop;
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Record_Global_Simulation_Time;

   --------------------------------
   -- Worst_Global_Simulation_Time --
   --------------------------------

   function Worst_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref)
      return Time
   is
   begin
      return Sim_Params_Lists.Item
        (The_Event,The_Result.Global_Simulation_Times).
        Worst_Simulation_Time;
   exception
      when List_Exceptions.Invalid_Index =>
         raise No_Results_For_Event;
   end Worst_Global_Simulation_Time;

   --------------------------------
   -- Worst_Global_Simulation_Time --
   --------------------------------

   function Worst_Global_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return Sim_Params_Lists.First_Item
        (The_Result.Global_Simulation_Times).Worst_Simulation_Time;
   exception
      when List_Exceptions.Empty =>
         raise No_Results_For_Event;
   end Worst_Global_Simulation_Time;

   -------------------------------
   -- Worst_Local_Simulation_Time --
   -------------------------------

   function Worst_Local_Simulation_Time
     (The_Result : Timing_Result)
      return Time
   is
   begin
      return The_Result.Local_Simulation_Time.Worst_Simulation_Time;
   end Worst_Local_Simulation_Time;

   -----------
   -- Print --
   -----------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out Time_Result_Lists.List;
      Indentation : Positive)
   is
      Names_Length : constant Integer := 16;
      The_Event : MAST.Events.Event_Ref;
      The_Time  : Time;
      Iterator : Time_Result_Lists.Iteration_Object;
   begin
      Set_Col(File,Ada.Text_IO.Count(Indentation));
      Put(File,'(');
      Time_Result_Lists.Rewind(The_List,Iterator);
      for I in 1..Time_Result_Lists.Size(The_List) loop
         if I>1 then
            MAST.IO.Print_Separator(File);
         end if;
         Time_Result_Lists.Get_Next_Item(The_Event,The_Time,The_List,
                                         Iterator);
         Set_Col(File,Ada.Text_IO.Count(Indentation+1));
         Put(File,'(');
         MAST.IO.Print_Arg
           (File,"Referenced_Event",
            IO.Name_Image(MAST.Events.Name(The_Event)),
            Indentation+2,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Time_Value",
            IO.Time_Image(The_Time),
            Indentation+2,Names_Length);
         Put(File,')');
      end loop;
      Put(File,')');
   end Print;

   -----------
   -- Print --
   -----------

   type Time_Parameter_Function is access function
     (Sim_Param : Simulation_Parameters) return String;

   function Worst_Simulation_Time
     (Sim_Param : Simulation_Parameters) return String is
   begin
      return IO.Time_Image(Sim_Param.Worst_Simulation_Time);
   end Worst_Simulation_Time;

   function Avg_Simulation_Time
     (Sim_Param : Simulation_Parameters) return String
   is
      Avg_Image : String(1..20);
   begin
      Put(Avg_Image,Float(Sim_Param.Sum_Of_Simulation_Times)/
          Float(Sim_Param.Num_Of_Simulation_Times),2,0);
      return Trim(Avg_Image,Both);
   end Avg_Simulation_Time;

   function Best_Simulation_Time
     (Sim_Param : Simulation_Parameters) return String is
   begin
      return IO.Time_Image(Sim_Param.Best_Simulation_Time);
   end Best_Simulation_Time;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out Sim_Params_Lists.List;
      Time_Value : Time_Parameter_Function;
      Indentation : Positive)
   is
      Names_Length : constant Integer := 16;
      The_Event : MAST.Events.Event_Ref;
      The_Params  : Simulation_Parameters;
      Iterator : Sim_Params_Lists.Iteration_Object;
   begin
      Set_Col(File,Ada.Text_IO.Count(Indentation));
      Put(File,'(');
      Sim_Params_Lists.Rewind(The_List,Iterator);
      for I in 1..Sim_Params_Lists.Size(The_List) loop
         if I>1 then
            MAST.IO.Print_Separator(File);
         end if;
         Sim_Params_Lists.Get_Next_Item(The_Event,The_Params,
                                        The_List,Iterator);
         Set_Col(File,Ada.Text_IO.Count(Indentation+1));
         Put(File,'(');
         MAST.IO.Print_Arg
           (File,"Referenced_Event",
            IO.Name_Image(MAST.Events.Name(The_Event)),
            Indentation+2,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Time_Value",
            Time_Value(The_Params),
            Indentation+2,Names_Length);
         Put(File,')');
      end loop;
      Put(File,')');
   end Print;

   -----------
   -- Print --
   -----------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out Miss_Ratio_Lists.List;
      Indentation : Positive)
   is
      Names_Length : constant Integer := 8;
      The_Deadline : Time;
      The_Ratio    : Percentage;
      The_Data     : Miss_Ratio_Data;
      Ratio_Image : String(1..20);
      Iterator : Miss_Ratio_Lists.Iteration_Object;
   begin
      Set_Col(File,Ada.Text_IO.Count(Indentation));
      Put(File,'(');
      Miss_Ratio_Lists.Rewind(The_List,Iterator);
      for I in 1..Miss_Ratio_Lists.Size(The_List) loop
         if I>1 then
            MAST.IO.Print_Separator(File);
         end if;
         Miss_Ratio_Lists.Get_Next_Item
           (The_Deadline,The_Data,The_List,Iterator);
         The_Ratio:=Float(The_Data.Num_Missed)/Float(The_Data.Total);
         Set_Col(File,Ada.Text_IO.Count(Indentation+1));
         Put(File,'(');
         MAST.IO.Print_Arg
           (File,"Deadline",
            IO.Time_Image(The_Deadline),
            Indentation+2,Names_Length);
         MAST.IO.Print_Separator(File);
         Put(Ratio_Image,The_Ratio,2,0);
         MAST.IO.Print_Arg
           (File,"Ratio",
            Trim(Ratio_Image,Both),
            Indentation+2,Names_Length);
         Put(File,')');
      end loop;
      Put(File,')');
   end Print;

   -----------
   -- Print --
   -----------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out List_Of_Miss_Lists.List;
      Indentation : Positive)
   is
      Names_Length : constant Integer := 16;
      The_Event : MAST.Events.Event_Ref;
      The_Miss_List  : List_Ref;
      Iterator : List_Of_Miss_Lists.Iteration_Object;
   begin
      Set_Col(File,Ada.Text_IO.Count(Indentation));
      Put(File,'(');
      List_Of_Miss_Lists.Rewind(The_List,Iterator);
      for I in 1..List_Of_Miss_Lists.Size(The_List) loop
         if I>1 then
            MAST.IO.Print_Separator(File);
         end if;
         List_Of_Miss_Lists.Get_Next_Item
           (The_Event,The_Miss_List,The_List,Iterator);
         Set_Col(File,Ada.Text_IO.Count(Indentation+1));
         Put(File,'(');
         MAST.IO.Print_Arg
           (File,"Referenced_Event",
            IO.Name_Image(MAST.Events.Name(The_Event)),
            Indentation+2,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Miss_Ratios","",Indentation+2,Names_Length);
         New_Line(File);
         Set_Col(File,Ada.Text_IO.Count(Indentation+4));
         Print(File,The_Miss_List.all,Indentation+5);
         Put(File,')');
      end loop;
      Put(File,')');
   end Print;

   -----------
   -- Print --
   -----------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Result;
      Indentation : Positive;
      Finalize    : Boolean:=False)
   is
   begin
      if Res.The_Link /= null then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Event_Name",
            IO.Name_Image(MAST.Graphs.Name(Res.The_Link)),
            Indentation,Names_Length);
      end if;
   end Print;

   -----------
   -- Print --
   -----------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Timing_Result;
      Indentation : Positive;
      Finalize    : Boolean:=False)
   is
   begin
      Ada.Text_IO.Set_Col(File,Ada.Text_IO.Count(Indentation));
      Ada.Text_IO.Put(File,"(");
      MAST.IO.Print_Arg
        (File,"Type",
         "Timing_Result",Indentation+1,Names_Length);
      Print(File,Result(Res),Indentation+1);
      if Res.Has_Worst_Local_Response_Time then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Worst_Local_Response_Time",
            IO.Time_Image(Res.Worst_Local_Response_Time),
            Indentation+1,Names_Length);
      end if;
      if Res.Has_Blocking_Time then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Num_Of_Suspensions",
            Natural'Image(Res.Num_Of_Suspensions),
            Indentation+1,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Blocking_Time",
            IO.Time_Image(Res.Blocking_Time),
            Indentation+1,Names_Length);
      end if;
      if Time_Result_Lists.Size(Res.Worst_Global_Response_Times)>0 and then
        Res.Has_Worst_Global_Response_Times
      then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Worst_Global_Response_Times","",
            Indentation+1,Names_Length);
         New_Line(File);
         Print(File,Res.Worst_Global_Response_Times,Indentation+4);
      end if;
      if Time_Result_Lists.Size(Res.Best_Global_Response_Times)>0 and then
        Res.Has_Best_Global_Response_Times
      then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Best_Global_Response_Times","",
            Indentation+1,Names_Length);
         New_Line(File);
         Print(File,Res.Best_Global_Response_Times,Indentation+4);
      end if;
      if Time_Result_Lists.Size(Res.Jitters)>0 and then
        Res.Has_Worst_Global_Response_Times
      then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Jitters","",
            Indentation+1,Names_Length);
         New_Line(File);
         Print(File,Res.Jitters,Indentation+4);
      end if;
      if Res.Has_Simulation_Times then
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Worst_Local_Simulation_Time",
            Worst_Simulation_Time(Res.Local_Simulation_Time),
            Indentation+1,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Avg_Local_Simulation_Time",
            Avg_Simulation_Time(Res.Local_Simulation_Time),
            Indentation+1,Names_Length);
         MAST.IO.Print_Separator(File);
         MAST.IO.Print_Arg
           (File,"Best_Local_Simulation_Time",
            Best_Simulation_Time(Res.Local_Simulation_Time),
            Indentation+1,Names_Length);
         if Sim_Params_Lists.Size(Res.Global_Simulation_Times)>0 then
            MAST.IO.Print_Separator(File);
            MAST.IO.Print_Arg
              (File,"Worst_Global_Simulation_Times","",
               Indentation+1,Names_Length);
            New_Line(File);
            Print(File,Res.Global_Simulation_Times,
                  Worst_Simulation_Time'access,Indentation+4);
            MAST.IO.Print_Separator(File);
            MAST.IO.Print_Arg
              (File,"Avg_Global_Simulation_Times","",
               Indentation+1,Names_Length);
            New_Line(File);
            Print(File,Res.Global_Simulation_Times,
                  Avg_Simulation_Time'access,Indentation+4);
            MAST.IO.Print_Separator(File);
            MAST.IO.Print_Arg
              (File,"Best_Global_Simulation_Times","",
               Indentation+1,Names_Length);
            New_Line(File);
            Print(File,Res.Global_Simulation_Times,
                  Best_Simulation_Time'access,Indentation+4);
         end if;
         if Miss_Ratio_Lists.Size(Res.Local_Miss_Ratios)>0 then
            MAST.IO.Print_Separator(File);
            MAST.IO.Print_Arg
              (File,"Local_Miss_Ratios","",
               Indentation+1,Names_Length);
            New_Line(File);
            Print(File,Res.Local_Miss_Ratios,Indentation+4);
         end if;
         Ada.Text_IO.Put(File,")");
         if List_Of_Miss_Lists.Size(Res.Global_Miss_Ratios)>0 then
            MAST.IO.Print_Separator(File);
            MAST.IO.Print_Arg
              (File,"Global_Miss_Ratios","",
               Indentation+1,Names_Length);
            New_Line(File);
            Print(File,Res.Global_Miss_Ratios,Indentation+4);
         end if;
      end if;
      Ada.Text_IO.Put(File,")");
   end Print;

   ------------------------
   -- Set_Num_Of_Suspensions
   ------------------------

   procedure Set_Num_Of_Suspensions
     (The_Result : in out Timing_Result;
      Num : Natural)
   is
   begin
      The_Result.Has_Blocking_Time:=True;
      The_Result.Num_Of_Suspensions:=Num;
   end Set_Num_Of_Suspensions;

   ------------------------
   -- Num_Of_Suspensions
   ------------------------

   function Num_Of_Suspensions
     (The_Result : Timing_Result) return Natural
   is
   begin
      return The_Result.Num_Of_Suspensions;
   end Num_Of_Suspensions;

end MAST.Results;
