-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with Ada.Text_IO,MAST.Graphs, MAST.Events, Hash_Lists;

package MAST.Results is

   use type MAST.Events.Event_Ref;

   No_Results_For_Event : exception;
   No_Results_For_Deadline : exception;

   type Time_Iteration_Object is private;
   type Sim_Iteration_Object is private;
   type Local_Ratio_Iteration_Object is private;
   type Global_Ratio_Iteration_Object is private;

   type Result is abstract tagged private;

   procedure Set_Link
     (The_Result : in out Result;
      The_Link : MAST.Graphs.Link_Ref);

   function Link_Of
     (The_Result : Result) return MAST.Graphs.Link_Ref;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Result;
      Indentation : Positive;
      Finalize    : Boolean:=False);

   type Result_Ref is access Result'Class;

   ------------------------
   -- Timing_Result
   ------------------------

   type Timing_Result is new Result with private;

   ----------------------------------------------------------------
   --               Worst Case Analysis Parameters               --
   ----------------------------------------------------------------

   type Timing_Parameters is record
      Worst_Response_Time : Time := Large_Time;
      Best_Response_Time : Time :=0.0;
      Jitter : Time := Large_Time;
   end record;

   ------------------------
   -- Local response times
   ------------------------
   procedure Set_Worst_Local_Response_Time
     (The_Result : in out Timing_Result;
      Worst_Local_Response_Time : Time);

   function Worst_Local_Response_Time
     (The_Result : Timing_Result) return Time;

   ------------------------
   -- Blocking
   ------------------------

   procedure Set_Blocking_Time
     (The_Result : in out Timing_Result;
      Blocking_Time : Time);

   function Blocking_Time
     (The_Result : Timing_Result) return Time;

   ------------------------
   -- Num_Of_Suspensions
   ------------------------

   procedure Set_Num_Of_Suspensions
     (The_Result : in out Timing_Result;
      Num : Natural);

   function Num_Of_Suspensions
     (The_Result : Timing_Result) return Natural;

   -----------------------------------
   -- Global response times and jitter
   -----------------------------------

   procedure Set_Worst_Global_Response_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Worst_Global_Response_Time : Time);
   procedure Set_Best_Global_Response_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Best_Global_Response_Time : Time);
   procedure Set_Jitter
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Jitter : Time);

   procedure Set_Worst_Global_Response_Time
     (The_Result : in out Timing_Result;
      Worst_Global_Response_Time : Time);
   -- sets data for first stored event
   -- may raise No_Results_For_Event
   procedure Set_Best_Global_Response_Time
     (The_Result : in out Timing_Result;
      Best_Global_Response_Time : Time);
   -- sets data for first stored event
   -- may raise No_Results_For_Event
   procedure Set_Jitter
     (The_Result : in out Timing_Result;
      Jitter : Time);
   -- sets data for first stored event
   -- may raise No_Results_For_Event

   function Worst_Global_Response_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event
   function Best_Global_Response_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event
   function Jitter
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event

   function Worst_Global_Response_Time
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event
   function Best_Global_Response_Time
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event
   function Jitter
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event



   procedure Rewind_Global_Timing_Parameters
     (The_Result : in Timing_Result;
      Iterator : out Time_Iteration_Object);
   procedure Get_Next_Global_Timing_Parameters
     (The_Result : in Timing_Result;
      The_Event : out MAST.Events.Event_Ref;
      Timing_Params : out Timing_Parameters;
      Iterator : in out Time_Iteration_Object);
   function Num_Of_Global_Timing_Parameters
     (The_Result : Timing_Result) return Natural;


   ---------------------------------------------------------------
   --                  Simulation Parameters                    --
   ---------------------------------------------------------------

   type Simulation_Parameters is record
      Worst_Simulation_Time   : Time:=0.0;
      Sum_Of_Simulation_Times : Time:=0.0;
      Best_Simulation_Time    : Time:=Large_Time;
      Num_Of_Simulation_Times : Natural:=0;
   end record;

   --------------------------
   -- Local simulation times
   --------------------------
   procedure Record_Local_Simulation_Time
     (The_Result : in out Timing_Result;
      Local_Simulation_Time : Time);
   -- affects worst, avg and best local simulation times,
   -- num of recorded local simulation times, and
   -- local miss ratios that have already been reset

   function Worst_Local_Simulation_Time
     (The_Result : Timing_Result) return Time;
   function Avg_Local_Simulation_Time
     (The_Result : Timing_Result) return Time;
   function Best_Local_Simulation_Time
     (The_Result : Timing_Result) return Time;

   -----------------------------------
   -- Global simulation times
   -----------------------------------

   procedure Record_Global_Simulation_Time
     (The_Result : in out Timing_Result;
      The_Event : MAST.Events.Event_Ref;
      Global_Simulation_Time : Time);
   -- if no global time had been recorded for the event,
   -- an entry is created for that event.
   -- affects worst, avg and best global simulation times,
   -- num of recorded global simulation times, and
   -- global miss ratios that have already been reset
   -- all, only for the specified event

   procedure Record_Global_Simulation_Time
     (The_Result : in out Timing_Result;
      Global_Simulation_Time : Time);
   -- same as above, but affects data for the first stored event
   -- may raise No_Results_For_Event

   function Worst_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event
   function Avg_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event
   function Best_Global_Simulation_Time
     (The_Result : Timing_Result;
      The_Event : MAST.Events.Event_Ref) return Time;
   -- may raise No_Results_For_Event

   function Worst_Global_Simulation_Time
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event
   function Avg_Global_Simulation_Time
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event
   function Best_Global_Simulation_Time
     (The_Result : Timing_Result) return Time;
   -- returns data for first stored event

   procedure Rewind_Global_Simulation_Times
     (The_Result : in Timing_Result;
      Iterator : out Sim_Iteration_Object);
   procedure Get_Next_Global_Simulation_Times
     (The_Result : in Timing_Result;
      The_Event : out MAST.Events.Event_Ref;
      Global_Simulation_Params : out Simulation_Parameters;
      Iterator : in out Sim_Iteration_Object);
   function Num_Of_Global_Simulation_Times
     (The_Result : Timing_Result) return Natural;

   ------------------------
   -- Local_Miss_Ratios
   ------------------------

   procedure Reset_Local_Miss_Ratio
     (The_Result : in out Timing_Result;
      The_Deadline : Time);
   -- Resets the entry for storing the miss ratio for the specified
   -- deadline, or creates one if it does not exist.
   -- The actual miss ratio is then set by successive calls to
   -- record_local_simulation_time

   procedure Rewind_Local_Miss_Ratios
     (The_Result : in Timing_Result;
      Iterator : out Local_Ratio_Iteration_Object);

   procedure Get_Next_Local_Miss_Ratio
     (The_Result : in Timing_Result;
      The_Deadline : out Time;
      The_Ratio : out Percentage;
      Iterator : in out LOcal_Ratio_Iteration_Object);

   function Local_Miss_Ratio
     (The_Result : Timing_Result;
      The_Deadline : Time) return Percentage;
   -- may raise No_Results_For_Deadline

   function Num_Of_Local_Miss_Ratios
     (The_Result : Timing_Result) return Natural;

   ------------------------
   -- Global_Miss_Ratios
   ------------------------

   procedure Reset_Global_Miss_Ratio
     (The_Result : in out Timing_Result;
      The_Deadline : Time;
      The_Event : MAST.Events.Event_Ref);
   -- Resets the entry for storing the miss ratio for the specified
   -- deadline and event, or creates one if it does not exist.
   -- The actual miss ratio is then set by successive calls to
   -- record_global_simulation_time, with the specified event

   procedure Rewind_Global_Miss_Ratios
     (The_Result : in Timing_Result;
      Iterator : out Global_Ratio_Iteration_Object);

   procedure Get_Next_Global_Miss_Ratio
     (The_Result : in Timing_Result;
      The_Deadline : out Time;
      The_Event : out MAST.Events.Event_Ref;
      The_Ratio : out Percentage;
      Iterator : in out Global_Ratio_Iteration_Object);

   function Global_Miss_Ratio
     (The_Result : Timing_Result;
      The_Deadline : Time;
      The_Event : MAST.Events.Event_Ref) return Percentage;
   -- may raise No_Results_For_Deadline or No_Results_For_Event

   function Num_Of_Global_Miss_Ratios
     (The_Result : Timing_Result) return Natural;

   -------------------
   -- Other operations
   -------------------

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Timing_Result;
      Indentation : Positive;
      Finalize    : Boolean:=False);

   function Create_Timing_Result
     (The_Link_Ref : MAST.Graphs.Link_Ref) return Result_Ref;
   -- creates a timing result bidirectionally linked to
   -- the specified arc

private

   type Result is tagged record
      The_Link : MAST.Graphs.Link_Ref;
   end record;

   type Miss_Ratio_Data is record
      Num_Missed, Total : Natural:=0;
   end record;

   package Time_Result_Lists is new Hash_Lists
     (Index => MAST.Events.Event_Ref,
      Element => Time, "=" => "=");

   package Sim_Params_Lists is new Hash_Lists
     (Index => MAST.Events.Event_Ref,
      Element => Simulation_Parameters, "=" => "=");

   package Miss_Ratio_Lists is new Hash_Lists
     (Index => Time, Element =>Miss_Ratio_Data, "=" => "=");

   type List_Ref is access Miss_Ratio_Lists.List;
   package List_Of_Miss_Lists is new Hash_Lists
     (Index => MAST.Events.Event_Ref,
      Element => List_Ref, "=" => "=");

   type Time_Iteration_Object is record
      Worst,Best,Jitters : Time_Result_Lists.Iteration_Object;
   end record;

   type Sim_Iteration_Object is new Sim_Params_Lists.Iteration_Object;

   type Local_Ratio_Iteration_Object is new
     Miss_Ratio_Lists.Iteration_Object;

   type Global_Ratio_Iteration_Object is record
      List_Of_Lists_Iterator : List_Of_Miss_Lists.Iteration_Object;
      Miss_List_Iterator : Miss_Ratio_Lists.Iteration_Object;
      Current_List : List_Ref;
      Current_Event : MAST.Events.Event_Ref;
   end record;



   type Timing_Result is new Result with record
      Has_Worst_Local_Response_Time,
      Has_Blocking_Time,
      Has_Worst_Global_Response_Times,
      Has_Best_Global_Response_Times,
      Has_Simulation_Times : Boolean:=False;
      Worst_Local_Response_Time : Time := 0.0;
      Blocking_Time : Time :=0.0;
      Num_Of_Suspensions : Natural:=0;
      Worst_Global_Response_Times,
      Best_Global_Response_Times,
      Jitters : Time_Result_Lists.List;
      Local_Simulation_Time : Simulation_Parameters;
      Global_Simulation_Times : Sim_Params_Lists.List;
      Local_Miss_Ratios : Miss_Ratio_Lists.List;
      Global_Miss_Ratios : List_Of_Miss_Lists.List;
   end record;

end MAST.Results;
