-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2000                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@ctr.unican.es                 --
--          Jose Javier Gutierrez  gutierjj@ctr.unican.es            --
--          Jose Carlos Palencia   palencij@ctr.unican.es            --
--          Jose Maria Drake       drakej@ctr.unican.es              --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with Ada.Text_IO,MAST.Graphs,MAST.Events, Mast.Operations,
  Named_Lists, Var_Strings;

package MAST.Transactions is

   Link_Not_Found : exception; -- raised by Find_*_Link

   Event_Not_Found : exception; -- raised by Find_Any_Event

   type Event_Handler_Iteration_Object is private;
   type Link_Iteration_Object is private;

   type Transaction is abstract tagged private;

   procedure Init
     (Trans : in out Transaction;
      Name : Var_Strings.Var_String);
   function Name (Trans : Transaction)
                  return Var_Strings.Var_String;

   procedure Add_Event_Handler
     (Trans : in out Transaction;
      The_Event_Handler : MAST.Graphs.Event_Handler_Ref);

   function Num_Of_Event_Handlers
     (Trans : Transaction )
      return Natural;

   procedure Rewind_Event_Handlers
     (Trans : in Transaction;
      Iterator : out Event_Handler_Iteration_Object);

   procedure Get_Next_Event_Handler
     (Trans : in Transaction;
      The_Event_Handler : out MAST.Graphs.Event_Handler_Ref;
      Iterator : in out Event_Handler_Iteration_Object);

   procedure Add_Internal_Event_Link
     (Trans : in out Transaction;
      The_Link : MAST.Graphs.Link_Ref);

   function Num_Of_Internal_Event_Links
     (Trans : Transaction )
      return Natural;

   function Find_Internal_Event_Link
     (Name : Var_Strings.Var_String;
      Trans : Transaction)
      return MAST.Graphs.Link_Ref;
   -- raises Link_Not_Found if link not found

   procedure Rewind_Internal_Event_Links
     (Trans : in Transaction;
      Iterator : out Link_Iteration_Object);

   procedure Get_Next_Internal_Event_Link
     (Trans : in Transaction;
      The_Link : out MAST.Graphs.Link_Ref;
      Iterator : in out Link_Iteration_Object);

   procedure Add_External_Event_Link
     (Trans : in out Transaction;
      The_Link : MAST.Graphs.Link_Ref);

   function Num_Of_External_Event_Links
     (Trans : Transaction )
      return Natural;

   function Find_External_Event_Link
     (Name : Var_Strings.Var_String;
      Trans : Transaction)
      return MAST.Graphs.Link_Ref;
   -- raises Link_Not_Found if link not found

   procedure Rewind_External_Event_Links
     (Trans : in Transaction;
      Iterator : out Link_Iteration_Object);

   procedure Get_Next_External_Event_Link
     (Trans : in Transaction;
      The_Link : out MAST.Graphs.Link_Ref;
      Iterator : in out Link_Iteration_Object);

   function Find_Any_Link
     (Name : Var_Strings.Var_String;
      Trans : Transaction)
      return MAST.Graphs.Link_Ref;
   -- raises Link_Not_Found if link not found

   function Find_Any_Event
     (Name : Var_Strings.Var_String;
      Trans : Transaction)
      return MAST.Events.Event_Ref;
   -- raises Event_Not_Found if link not found

   procedure Scale
     (Trans : Transaction; Factor : Normalized_Execution_Time);
   -- Scales the execution times of all operations in the transaction

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Transaction;
      Indentation : Positive;
      Finalize    : Boolean:=False) is abstract;

   procedure Print_Results
     (File : Ada.Text_IO.File_Type;
      Res : in out Transaction;
      Indentation : Positive;
      Finalize    : Boolean:=False);

   type Transaction_Ref is access Transaction'Class;

   function Clone
     (Trans : Transaction)
     return Transaction_Ref is abstract;

   procedure Adjust_Operations
     (Trans : in out Transaction;
      Op_List : Operations.Lists.List);
   -- adjusts internal pointers referencing operations; used after clonning

   function Name (Trans_Ref : Transaction_Ref )
                  return Var_Strings.Var_String;

   package Lists is new Named_Lists
     (Element => Transaction_Ref,
      Name    => Name);

   procedure Print
     (File : Ada.Text_IO.File_Type;
      The_List : in out Lists.List;
      Indentation : Positive);

   function Clone
     (The_List : Lists.List)
     return Lists.List;

   procedure Adjust_Operations
     (The_List : Lists.List;
      Op_List : Operations.Lists.List);
   -- adjusts internal pointers referencing operations; used after clonning

   procedure Print_Results
     (File : Ada.Text_IO.File_Type;
      The_List : in out Lists.List;
      Indentation : Positive);

   type Regular_Transaction is new Transaction with private;

   procedure Print
     (File : Ada.Text_IO.File_Type;
      Res : in out Regular_Transaction;
      Indentation : Positive;
      Finalize    : Boolean:=False);

   function Clone
     (Trans : Regular_Transaction)
     return Transaction_Ref;


private

      type Transaction is abstract tagged record
         Name : Var_Strings.Var_String;
         Graph : MAST.Graphs.Graph;
      end record;

      type Event_Handler_Iteration_Object is new
        Graphs.Event_Handler_Lists.Index;
      type Link_Iteration_Object is new
        Graphs.Link_Lists.Index;

      type Regular_Transaction is new Transaction with null record;

end MAST.Transactions;
