-----------------------------------------------------------------------
--                              Mast                                 --
--     Modelling and Analysis Suite for Real-Time Applications       --
--                                                                   --
--                       Copyright (C) 2001                          --
--                 Universidad de Cantabria, SPAIN                   --
--                                                                   --
-- Authors: Michael Gonzalez       mgh@unican.es                     --
--          Jose Javier Gutierrez  gutierjj@unican.es                --
--          Jose Carlos Palencia   palencij@unican.es                --
--          Jose Maria Drake       drakej@unican.es                  --
--                                                                   --
-- This program is free software; you can redistribute it and/or     --
-- modify it under the terms of the GNU General Public               --
-- License as published by the Free Software Foundation; either      --
-- version 2 of the License, or (at your option) any later version.  --
--                                                                   --
-- This program is distributed in the hope that it will be useful,   --
-- but WITHOUT ANY WARRANTY; without even the implied warranty of    --
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU --
-- General Public License for more details.                          --
--                                                                   --
-- You should have received a copy of the GNU General Public         --
-- License along with this program; if not, write to the             --
-- Free Software Foundation, Inc., 59 Temple Place - Suite 330,      --
-- Boston, MA 02111-1307, USA.                                       --
--                                                                   --
-----------------------------------------------------------------------

with Ada.Text_IO,Ada.Command_Line,Ada.Exceptions,
  MAST_Parser,MAST.Systems,
  MAST.Consistency_Checks,Mast.Tools, Mast.IO,
  Mast.Transactions.Results,Mast.Timing_Requirements,
  Mast.Linear_Priority_Assignment_Tools,
  Mast.HOPA_Parameters,
  Mast.Annealing_Parameters,
  Mast.Restrictions,
  Mast.Tool_Exceptions,Mast_Parser_Error_Report,Var_Strings,
  Mast_Analysis_Help;
use Mast,Ada.Text_IO,Ada.Command_Line,Ada.Exceptions,Var_strings;
use type Mast.Tools.Worst_Case_Analysis_Tool;
use type Mast.Tools.Priority_Assignment_Tool;
procedure Mast_Analysis
is

   Version_String : constant String:="1.1";

   The_Tool : Tools.Worst_Case_Analysis_Tool;
   The_Priority_Assignment_Tool : Tools.Priority_Assignment_Tool;

   The_System : MAST.Systems.System;
   Input_File, Output_File, Description_File : File_Type;
   type Option is (Verbose,Calculate_Ceilings,Assign_Priorities,
                   Assignment_Parameters, Dump_System_Description,
                   Calculate_Slacks, Check_Deadlocks);
   type Priority_Assignment_Technique is
     (Default,HOPA,Annealing,Monoprocessor);
   Flag : array(Option) of Boolean:=
     (Verbose                 => False,
      Calculate_Ceilings      => False,
      Assign_Priorities       => False,
      Assignment_Parameters   => False,
      Dump_System_Description => False,
      Calculate_Slacks        => False,
      Check_Deadlocks         => False);
   Technique : Priority_Assignment_Technique:=Default;
   Parameters_Filename : Var_String:=To_Var_String
     ("priority_assignment_parameters.txt");
   Description_Name : Var_String;
   Output_File_Specified : Boolean:=False;
   Arg, Input_File_Arg : Natural;
   System_Slack : Float;
   Slack_List : Transactions.Results.Results_List;
   Wrong_Format,Unrecognizable : exception;

   procedure Close_Files is
   begin
      Close(Input_File);
      if Output_File_Specified then
         Close(Output_File);
         Ada.Text_IO.Set_Output(Standard_Output);
      end if;
   end Close_Files;


begin
   if Argument_Count=0 then
      raise Wrong_Format;
   elsif Argument_Count=1 then
      if (Argument(1)="-h" or else Argument(1)="-help")
      then
         Mast_Analysis_Help;
      else
         raise Wrong_Format;
      end if;
   else
      if Argument(1)="parse" then
         null;
      elsif Argument(1)="holistic" then
         The_Tool:=Tools.Holistic_Analysis'Access;
      elsif Argument(1)="offset_based_unoptimized" then
         The_Tool:=Tools.Offset_Based_Unoptimized_Analysis'Access;
      elsif Argument(1)="offset_based" then
         The_Tool:=Tools.Offset_Based_Analysis'Access;
      else
         raise Wrong_Format;
      end if;
      Arg:=2;
      while Arg < Argument_Count loop
         if Argument(Arg)="-v" or else Argument(Arg)="-verbose" then
            Flag(Verbose):=True;
         elsif Argument(Arg)="-c" or else Argument(Arg)="-ceilings" then
            Flag(Calculate_Ceilings):=True;
         elsif Argument(Arg)="-p" or else Argument(Arg)="-priorities" then
            Flag(Assign_Priorities):=True;
            Flag(Calculate_Ceilings):=True;
         elsif Argument(Arg)="-t" or else Argument(Arg)="-technique" then
            Arg:=Arg+1;
            if Argument(Arg)="hopa" then
               Technique:=Hopa;
               The_Priority_Assignment_Tool :=Tools.Linear_Hopa'Access;
            elsif Argument(Arg)="annealing" then
               Technique:=Annealing;
               The_Priority_Assignment_Tool :=
                  Tools.Linear_Simulated_Annealing_Priority_Assignment'Access;
            elsif Argument(Arg)="monoprocessor" then
               Technique:=Monoprocessor;
               The_Priority_Assignment_Tool :=
                 Tools.Monoprocessor_Priority_Assignment'Access;
            else
               raise Wrong_Format;
            end if;
         elsif Argument(Arg)="-a" or else
           Argument(Arg)="-assignment_parameters"
         then
            Flag(Assignment_Parameters):=True;
            Arg:=Arg+1;
            Parameters_Filename :=To_Var_String(Argument(Arg));
         elsif Argument(Arg)="-o" or else Argument(Arg)="-ordering" then
            Flag(Check_Deadlocks):=True;
         elsif Argument(Arg)="-d" or else Argument(Arg)="-description" then
            Flag(Dump_System_Description):=True;
            Arg:=Arg+1;
            Description_Name :=To_Var_String(Argument(Arg));
         elsif Argument(Arg)="-s" or else Argument(Arg)="-slack" then
            Flag(Calculate_Slacks):=True;
         else
            if Arg=Argument_Count-1 then
               Output_File_Specified:=True;
            else
               raise Wrong_Format;
            end if;
         end if;
         Arg:=Arg+1;
      end loop;
      begin
         if Output_File_Specified then
            Input_File_Arg:=Argument_Count-1;
            Ada.Text_IO.Create(Output_File,Out_File,Argument(Argument_Count));
            Ada.Text_IO.Set_Output(Output_File);
         else
            Input_File_Arg:=Argument_Count;
         end if;
         Ada.Text_IO.Open(Input_File,In_File,Argument(Input_File_Arg));
         Ada.Text_IO.Set_Input(Input_File);
         Put_Line("MAST Version: "&Version_String);
         New_Line;
         Put_Line("Parsing input file");
         begin
             MAST_Parser(The_System);
         exception
             when Constraint_Error =>
                 raise Unrecognizable;
         end;
         Ada.Text_IO.Set_Input(Standard_Input);
         if consistency_checks.Consistent_Transaction_Graphs(The_System)
         then
            if Flag(Verbose)then
               Put_Line("Consistent_Transaction_Graphs met");
            end if;
         else
            Tool_Exceptions.Set_Restriction_Message
              ("Consistent_Transaction_Graphs no met");
            raise Tool_Exceptions.Restriction_Not_Met;
         end if;

         if consistency_checks.Consistent_Shared_Resource_Usage(The_System)
         then
            if Flag(Verbose) then
               Put_Line("Consistent_Shared_Resource_Usage met");
            end if;
         else
            Tool_Exceptions.Set_Restriction_Message
              ("Consistent_Shared_Resource_Usage no met");
            raise Tool_Exceptions.Restriction_Not_Met;
         end if;

         if Flag(Assign_Priorities) then
            if Technique = Default then
               if Restrictions.Monoprocessor_Only(The_System,False) and then
                 Restrictions.Simple_Transactions_Only(The_System,False)
               then
                  Technique:=Monoprocessor;
                  The_Priority_Assignment_Tool:=
                     Tools.Monoprocessor_Priority_Assignment'Access;
               else
                  Technique:=Hopa;
                  The_Priority_Assignment_Tool:=
                     Tools.Linear_Hopa'Access;
               end if;
            end if;
            begin
               case Technique is
                  when Hopa =>
                     Hopa_Parameters.
                       Load_Parameters(To_String(Parameters_Filename));
                  when Annealing =>
                     Annealing_Parameters.
                       Load_Parameters(To_String(Parameters_Filename));
                  when others => null;
               end case;
            exception
               when Tool_Exceptions.Invalid_Format =>
                  Tool_Exceptions.Set_Tool_Failure_Message
                    ("Invalid Format in priority assignment parameters file");
                  raise Tool_Exceptions.Tool_Failure;
               when Tool_Exceptions.File_Does_Not_Exist =>
                  if Flag(Assignment_Parameters) then
                     Tool_Exceptions.Set_Tool_Failure_Message
                       ("Priority assignment parameters file does not exist");
                     raise Tool_Exceptions.Tool_Failure;
                  end if;
            end;
            if (Technique = HOPA) or
              (Technique = Annealing) or
              (Technique = Monoprocessor) then
               if The_Tool /= null then
                  Put_Line("Invoking the priority assignment tool...");
                  The_Priority_Assignment_Tool
                    (The_System,The_Tool,Flag(Verbose));
               end if;
            end if;
         end if;

         if Flag(Calculate_Ceilings) then
            Put_Line("Calculating Ceilings");
            Tools.Calculate_Ceilings_For_PCP_Resources
              (The_System,Flag(Verbose));
         end if;

         if Flag(Dump_System_Description) then
            Create(Description_File,Out_File,To_String(Description_Name));
            Mast.Systems.Print(Description_File,The_System);
            Close(Description_File);
         end if;

         if Flag(Check_Deadlocks) then
            Put_Line("----------------------------------------------------");
            Put_Line("Total ordering check for resources not yet available");
            Put_Line("----------------------------------------------------");
         end if;

         if The_Tool/=null then
            Put_Line("Invoking the analysis tool...");
            The_Tool(The_System,Flag(Verbose));
            Mast.Systems.Print_Results(Current_Output,The_System);

            if Flag(Calculate_Slacks) then
               begin
                  Put_Line("Calculating Slacks...");
                  Mast.Tools.Calculate_Transaction_Slacks
                    (The_System,The_Tool,Slack_List,Flag(Verbose));
                  Mast.Transactions.Results.Print(Current_Output,Slack_List);
                  Mast.Tools.Calculate_System_Slack
                    (The_System,The_Tool,System_Slack,Flag(Verbose));
                  Ada.Text_IO.Put_Line
                    ("System slack : "&IO.Float_Image(System_Slack*100.0)&"%");
               exception
                  when Timing_Requirements.Inconclusive =>
                     Put_Line ("Inconclusive results when calculating slacks");
                     Put_Line ("Worst Case Analysis tools cannot check "&
                               "soft deadlines");
               end;
            end if;
         end if;

         Close_Files;
      end;
   end if;

exception

   when Wrong_Format =>
      Put_Line("Usage:");
      Put_Line("    mast_analysis -help");
      Put_Line("    mast_analysis tool_name [options] input_file "&
               "[output_file]");
      Put_Line("        tool_name:");
      Put_Line("          parse");
      Put_Line("          holistic");
      Put_Line("          linear_unoptimized");
      Put_Line("          linear");
      New_Line;
      Put_Line("        options:");
      Put_Line("          -v, -verbose");
      Put_Line("          -c, -ceilings");
      Put_Line("          -p, -priorities");
      Put_Line("          -t name, -technique name");
      Put_Line("               name can be :hopa,annealing,monoprocessor");
      Put_Line("          -o, -ordering");
      Put_Line("          -d filename, -description filename");
      Put_Line("          -s, -slack");
      Close_Files;

   when Name_Error |Status_Error | Use_error=>
      Put_Line("Input file not found");
      Close_Files;

   when Unrecognizable | End_Error | Layout_Error =>
      Put_Line("Input file has unrecognizable format");
      Close_Files;

   when Tool_Exceptions.Tool_Failure =>
      Put_Line("Tool Failure exception");
      Put_Line(Tool_Exceptions.Tool_Failure_Message);
      Close_Files;

   when Tool_Exceptions.Restriction_Not_Met =>
      Put_Line("Restriction Not Met");
      Put_Line(Tool_Exceptions.Restriction_Message);
      Close_Files;

   when Mast_Parser_Error_Report.Syntax_Error =>
      Close_Files;

   when Other_Exception : others =>
      Put_Line ("------------------Mast Internal Error ----------------");
      Put_Line ("-- Unexpected exception : "&
                Exception_Name(Other_Exception));
      Put_Line ("--");
      Put_Line ("-- Please send e-mail to mgh@unican.es");
      Put_Line ("-- including the following information:");
      Put_Line ("--      . Exception name obtained");
      Put_Line ("--      . Command line with the options that were used");
      Put_Line ("--      . The input file as an e-mail attachment");
      Put_Line ("--      . The MAST version used");
      Put_Line ("-- we will try to contact you as soon as possible");
      Put_Line ("-- The MAST team");
      Put_Line ("------------------------------------------------------");
      Close_Files;

end Mast_Analysis;
